"use strict";

var _test = require("test");
var _assert = require("assert");
var _sinon = _interopRequireDefault(require("sinon"));
var _encryptor = require("./_encryptor");
var _promiseToolbox = require("promise-toolbox");
var _ = require(".");
var _abstract = _interopRequireDefault(require("./abstract"));
var _fsExtra = _interopRequireDefault(require("fs-extra"));
var _rimraf = _interopRequireDefault(require("rimraf"));
var _tmp = _interopRequireDefault(require("tmp"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const TIMEOUT = 10e3;
class TestHandler extends _abstract.default {
  constructor(impl) {
    super({
      url: 'test://'
    }, {
      timeout: TIMEOUT
    });
    Object.defineProperty(this, 'isEncrypted', {
      get: () => false
    });
    Object.keys(impl).forEach(method => {
      this[`_${method}`] = impl[method];
    });
  }
}
const noop = Function.prototype;
const clock = _sinon.default.useFakeTimers();
(0, _test.describe)('closeFile()', () => {
  (0, _test.it)(`throws in case of timeout`, async () => {
    const testHandler = new TestHandler({
      closeFile: () => new Promise(() => {})
    });
    const promise = testHandler.closeFile({
      fd: undefined,
      path: ''
    });
    clock.tick(TIMEOUT);
    await _assert.strict.rejects(promise, _promiseToolbox.TimeoutError);
  });
});
(0, _test.describe)('getInfo()', () => {
  (0, _test.it)('throws in case of timeout', async () => {
    const testHandler = new TestHandler({
      getInfo: () => new Promise(() => {})
    });
    const promise = testHandler.getInfo();
    clock.tick(TIMEOUT);
    await _assert.strict.rejects(promise, _promiseToolbox.TimeoutError);
  });
});
(0, _test.describe)('getSize()', () => {
  (0, _test.it)(`throws in case of timeout`, async () => {
    const testHandler = new TestHandler({
      getSize: () => new Promise(() => {})
    });
    const promise = testHandler.getSize('');
    clock.tick(TIMEOUT);
    await _assert.strict.rejects(promise, _promiseToolbox.TimeoutError);
  });
});
(0, _test.describe)('list()', () => {
  (0, _test.it)(`throws in case of timeout`, async () => {
    const testHandler = new TestHandler({
      list: () => new Promise(() => {})
    });
    const promise = testHandler.list('.');
    clock.tick(TIMEOUT);
    await _assert.strict.rejects(promise, _promiseToolbox.TimeoutError);
  });
});
(0, _test.describe)('openFile()', () => {
  (0, _test.it)(`throws in case of timeout`, async () => {
    const testHandler = new TestHandler({
      openFile: () => new Promise(() => {})
    });
    const promise = testHandler.openFile('path');
    clock.tick(TIMEOUT);
    await _assert.strict.rejects(promise, _promiseToolbox.TimeoutError);
  });
});
(0, _test.describe)('rename()', () => {
  (0, _test.it)(`throws in case of timeout`, async () => {
    const testHandler = new TestHandler({
      rename: () => new Promise(() => {})
    });
    const promise = testHandler.rename('oldPath', 'newPath');
    clock.tick(TIMEOUT);
    await _assert.strict.rejects(promise, _promiseToolbox.TimeoutError);
  });
});
(0, _test.describe)('rmdir()', () => {
  (0, _test.it)(`throws in case of timeout`, async () => {
    const testHandler = new TestHandler({
      rmdir: () => new Promise(() => {})
    });
    const promise = testHandler.rmdir('dir');
    clock.tick(TIMEOUT);
    await _assert.strict.rejects(promise, _promiseToolbox.TimeoutError);
  });
});
(0, _test.describe)('encryption', () => {
  let dir;
  (0, _test.beforeEach)(async () => {
    dir = await (0, _promiseToolbox.pFromCallback)(cb => _tmp.default.dir(cb));
  });
  (0, _test.after)(async () => {
    await (0, _rimraf.default)(dir);
  });
  (0, _test.it)('sync should NOT create metadata if missing (not encrypted)', async () => {
    await _promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}`
    }), noop);
    _assert.strict.deepEqual(await _fsExtra.default.readdir(dir), []);
  });
  (0, _test.it)('sync should create metadata if missing (encrypted)', async () => {
    await _promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}?encryptionKey="73c1838d7d8a6088ca2317fb5f29cd00"`
    }), noop);
    _assert.strict.deepEqual(await _fsExtra.default.readdir(dir), ['encryption.json', 'metadata.json']);
    const encryption = JSON.parse(await _fsExtra.default.readFile(`${dir}/encryption.json`, 'utf-8'));
    _assert.strict.equal(encryption.algorithm, _encryptor.DEFAULT_ENCRYPTION_ALGORITHM);
    _assert.strict.rejects(async () => JSON.parse(await _fsExtra.default.readFile(`${dir}/metadata.json`)));
  });
  (0, _test.it)('sync should not modify existing metadata', async () => {
    await _fsExtra.default.writeFile(`${dir}/encryption.json`, `{"algorithm": "none"}`);
    await _fsExtra.default.writeFile(`${dir}/metadata.json`, `{"random": "NOTSORANDOM"}`);
    await _promiseToolbox.Disposable.use(await (0, _.getSyncedHandler)({
      url: `file://${dir}`
    }), noop);
    const encryption = JSON.parse(await _fsExtra.default.readFile(`${dir}/encryption.json`, 'utf-8'));
    _assert.strict.equal(encryption.algorithm, 'none');
    const metadata = JSON.parse(await _fsExtra.default.readFile(`${dir}/metadata.json`, 'utf-8'));
    _assert.strict.equal(metadata.random, 'NOTSORANDOM');
  });
  (0, _test.it)('should modify metadata if empty', async () => {
    await _promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}`
    }), noop);
    await _promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}?encryptionKey="73c1838d7d8a6088ca2317fb5f29cd00"`
    }), noop);
    let encryption = JSON.parse(await _fsExtra.default.readFile(`${dir}/encryption.json`, 'utf-8'));
    _assert.strict.equal(encryption.algorithm, _encryptor.DEFAULT_ENCRYPTION_ALGORITHM);
    await _promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}`
    }), noop);
    encryption = JSON.parse(await _fsExtra.default.readFile(`${dir}/encryption.json`, 'utf-8'));
    _assert.strict.equal(encryption.algorithm, 'none');
  });
  (0, _test.it)('sync should work with encrypted', _promiseToolbox.Disposable.wrap(async function* () {
    const encryptor = (0, _encryptor._getEncryptor)(_encryptor.DEFAULT_ENCRYPTION_ALGORITHM, '73c1838d7d8a6088ca2317fb5f29cd91');
    await _fsExtra.default.writeFile(`${dir}/encryption.json`, `{"algorithm": "${_encryptor.DEFAULT_ENCRYPTION_ALGORITHM}"}`);
    await _fsExtra.default.writeFile(`${dir}/metadata.json`, encryptor.encryptData(`{"random": "NOTSORANDOM"}`));
    const handler = yield (0, _.getSyncedHandler)({
      url: `file://${dir}?encryptionKey="73c1838d7d8a6088ca2317fb5f29cd91"`
    });
    const encryption = JSON.parse(await _fsExtra.default.readFile(`${dir}/encryption.json`, 'utf-8'));
    _assert.strict.equal(encryption.algorithm, _encryptor.DEFAULT_ENCRYPTION_ALGORITHM);
    const metadata = JSON.parse(await handler.readFile(`./metadata.json`));
    _assert.strict.equal(metadata.random, 'NOTSORANDOM');
  }));
  (0, _test.it)('sync should fail when changing key on non empty remote ', async () => {
    const encryptor = (0, _encryptor._getEncryptor)(_encryptor.DEFAULT_ENCRYPTION_ALGORITHM, '73c1838d7d8a6088ca2317fb5f29cd91');
    await _fsExtra.default.writeFile(`${dir}/encryption.json`, `{"algorithm": "${_encryptor.DEFAULT_ENCRYPTION_ALGORITHM}"}`);
    await _fsExtra.default.writeFile(`${dir}/metadata.json`, encryptor.encryptData(`{"random": "NOTSORANDOM"}`));
    await _promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}?encryptionKey="73c1838d7d8a6088ca2317fb5f29cd00"`
    }), noop);
    await _fsExtra.default.writeFile(`${dir}/nonempty.json`, 'content');
    await _assert.strict.rejects(_promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}?encryptionKey="73c1838d7d8a6088ca2317fb5f29cd10"`
    }), noop));
  });
  (0, _test.it)('sync should fail when changing algorithm', async () => {
    const encryptor = (0, _encryptor._getEncryptor)('aes-256-cbc', '73c1838d7d8a6088ca2317fb5f29cd91');
    await _fsExtra.default.writeFile(`${dir}/encryption.json`, `{"algorithm": "aes-256-gmc"}`);
    await _fsExtra.default.writeFile(`${dir}/metadata.json`, encryptor.encryptData(`{"random": "NOTSORANDOM"}`));
    await _fsExtra.default.writeFile(`${dir}/nonempty.json`, 'content');
    await _assert.strict.rejects(_promiseToolbox.Disposable.use((0, _.getSyncedHandler)({
      url: `file://${dir}?encryptionKey="73c1838d7d8a6088ca2317fb5f29cd91"`
    }), noop));
  });
});
//# sourceMappingURL=abstract.test.js.map