"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.validChecksumOfReadStream = exports.createChecksumStream = void 0;
var _through = _interopRequireDefault(require("through2"));
var _crypto = require("crypto");
var _promiseToolbox = require("promise-toolbox");
var _invert = _interopRequireDefault(require("lodash/invert.js"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const ALGORITHM_TO_ID = {
  md5: '1',
  sha256: '5',
  sha512: '6'
};
const ID_TO_ALGORITHM = (0, _invert.default)(ALGORITHM_TO_ID);
const createChecksumStream = (algorithm = 'md5') => {
  const algorithmId = ALGORITHM_TO_ID[algorithm];
  if (!algorithmId) {
    throw new Error(`unknown algorithm: ${algorithm}`);
  }
  const hash = (0, _crypto.createHash)(algorithm);
  const {
    promise,
    resolve,
    reject
  } = (0, _promiseToolbox.defer)();
  const stream = (0, _through.default)((chunk, enc, callback) => {
    hash.update(chunk);
    callback(null, chunk);
  }, callback => {
    resolve(`$${algorithmId}$$${hash.digest('hex')}`);
    callback();
  }).once('error', reject);
  stream.checksum = promise;
  return stream;
};
exports.createChecksumStream = createChecksumStream;
const validChecksumOfReadStream = (stream, expectedChecksum) => {
  const algorithmId = expectedChecksum.slice(1, expectedChecksum.indexOf('$', 1));
  if (!algorithmId) {
    throw new Error(`unknown algorithm: ${algorithmId}`);
  }
  const hash = (0, _crypto.createHash)(ID_TO_ALGORITHM[algorithmId]);
  const wrapper = stream.pipe((0, _through.default)({
    highWaterMark: 0
  }, (chunk, enc, callback) => {
    hash.update(chunk);
    callback(null, chunk);
  }, callback => {
    const checksum = `$${algorithmId}$$${hash.digest('hex')}`;
    callback(checksum !== expectedChecksum ? new Error(`Bad checksum (${checksum}), expected: ${expectedChecksum}`) : null);
  }));
  stream.on('error', error => wrapper.emit('error', error));
  wrapper.checksumVerified = (0, _promiseToolbox.fromEvent)(wrapper, 'end');
  return wrapper;
};
exports.validChecksumOfReadStream = validChecksumOfReadStream;
//# sourceMappingURL=checksum.js.map