"use strict";

var _test = require("test");
var _assert = require("assert");
require("dotenv/config");
var _lodash = require("lodash");
var _os = require("os");
var _ = require(".");
const unsecureRandomBytes = n => {
  const bytes = Buffer.alloc(n);
  const odd = n & 1;
  for (let i = 0, m = n - odd; i < m; i += 2) {
    bytes.writeUInt16BE(Math.random() * 65536 | 0, i);
  }
  if (odd) {
    bytes.writeUInt8(Math.random() * 256 | 0, n - 1);
  }
  return bytes;
};
const TEST_DATA_LEN = 1024;
const TEST_DATA = unsecureRandomBytes(TEST_DATA_LEN);
const rejectionOf = p => p.then(value => {
  throw value;
}, reason => reason);
const handlers = [`file://${(0, _os.tmpdir)()}`];
if (process.env.xo_fs_nfs) handlers.push(process.env.xo_fs_nfs);
if (process.env.xo_fs_smb) handlers.push(process.env.xo_fs_smb);
handlers.forEach(url => {
  (0, _test.describe)(url, () => {
    let handler;
    const testWithFileDescriptor = (path, flags, fn) => {
      (0, _test.it)('with path', () => fn({
        file: path,
        flags
      }));
      (0, _test.it)('with file descriptor', async () => {
        const file = await handler.openFile(path, flags);
        try {
          await fn({
            file
          });
        } finally {
          await handler.closeFile(file);
        }
      });
    };
    (0, _test.before)(async () => {
      handler = (0, _.getHandler)({
        url
      }).addPrefix(`xo-fs-tests-${Date.now()}`);
      await handler.sync();
    });
    (0, _test.after)(async () => {
      await handler.forget();
      handler = undefined;
    });
    (0, _test.beforeEach)(async () => {
      await handler.mkdir('.');
    });
    (0, _test.afterEach)(async () => {
      await handler.rmtree('.');
    });
    (0, _test.describe)('#type', () => {
      (0, _test.it)('returns the type of the remote', () => {
        _assert.strict.equal(typeof handler.type, 'string');
      });
    });
    (0, _test.describe)('#getInfo()', () => {
      let info;
      (0, _test.before)(async () => {
        info = await handler.getInfo();
      });
      (0, _test.it)('should return an object with info', async () => {
        _assert.strict.equal(typeof info, 'object');
      });
      (0, _test.it)('should return correct type of attribute', async () => {
        if (info.size !== undefined) {
          _assert.strict.equal(typeof info.size, 'number');
        }
        if (info.used !== undefined) {
          _assert.strict.equal(typeof info.used, 'number');
        }
      });
    });
    (0, _test.describe)('#getSize()', () => {
      (0, _test.before)(() => handler.outputFile('file', TEST_DATA));
      testWithFileDescriptor('file', 'r', async () => {
        _assert.strict.equal(await handler.getSize('file'), TEST_DATA_LEN);
      });
    });
    (0, _test.describe)('#list()', () => {
      (0, _test.it)(`should list the content of folder`, async () => {
        await handler.outputFile('file', TEST_DATA);
        _assert.strict.deepEqual(await handler.list('.'), ['file']);
      });
      (0, _test.it)('can prepend the directory to entries', async () => {
        await handler.outputFile('dir/file', '');
        _assert.strict.deepEqual(await handler.list('dir', {
          prependDir: true
        }), ['/dir/file']);
      });
      (0, _test.it)('throws ENOENT if no such directory', async () => {
        await handler.rmtree('dir');
        _assert.strict.equal((await rejectionOf(handler.list('dir'))).code, 'ENOENT');
      });
      (0, _test.it)('can returns empty for missing directory', async () => {
        _assert.strict.deepEqual(await handler.list('dir', {
          ignoreMissing: true
        }), []);
      });
    });
    (0, _test.describe)('#mkdir()', () => {
      (0, _test.it)('creates a directory', async () => {
        await handler.mkdir('dir');
        _assert.strict.deepEqual(await handler.list('.'), ['dir']);
      });
      (0, _test.it)('does not throw on existing directory', async () => {
        await handler.mkdir('dir');
        await handler.mkdir('dir');
      });
      (0, _test.it)('throws ENOTDIR on existing file', async () => {
        await handler.outputFile('file', '');
        const error = await rejectionOf(handler.mkdir('file'));
        _assert.strict.equal(error.code, 'ENOTDIR');
      });
    });
    (0, _test.describe)('#mktree()', () => {
      (0, _test.it)('creates a tree of directories', async () => {
        await handler.mktree('dir/dir');
        _assert.strict.deepEqual(await handler.list('.'), ['dir']);
        _assert.strict.deepEqual(await handler.list('dir'), ['dir']);
      });
      (0, _test.it)('does not throw on existing directory', async () => {
        await handler.mktree('dir/dir');
        await handler.mktree('dir/dir');
      });
      (0, _test.it)('throws ENOTDIR on existing file', async () => {
        await handler.outputFile('dir/file', '');
        const error = await rejectionOf(handler.mktree('dir/file'));
        _assert.strict.equal(error.code, 'ENOTDIR');
      });
      (0, _test.it)('throws ENOTDIR on existing file in path', async () => {
        await handler.outputFile('file', '');
        const error = await rejectionOf(handler.mktree('file/dir'));
        _assert.strict.equal(error.code, 'ENOTDIR');
      });
    });
    (0, _test.describe)('#outputFile()', () => {
      (0, _test.it)('writes data to a file', async () => {
        await handler.outputFile('file', TEST_DATA);
        _assert.strict.deepEqual(await handler.readFile('file'), TEST_DATA);
      });
      (0, _test.it)('throws on existing files', async () => {
        await handler.unlink('file');
        await handler.outputFile('file', '');
        const error = await rejectionOf(handler.outputFile('file', ''));
        _assert.strict.equal(error.code, 'EEXIST');
      });
      (0, _test.it)("shouldn't timeout in case of the respect of the parallel execution restriction", async () => {
        const handler = (0, _.getHandler)({
          url
        }, {
          maxParallelOperations: 1
        });
        await handler.sync();
        await handler.outputFile(`xo-fs-tests-${Date.now()}/test`, '');
      }, 40);
    });
    (0, _test.describe)('#read()', () => {
      (0, _test.before)(() => handler.outputFile('file', TEST_DATA));
      const start = (0, _lodash.random)(TEST_DATA_LEN);
      const size = (0, _lodash.random)(TEST_DATA_LEN);
      testWithFileDescriptor('file', 'r', async ({
        file
      }) => {
        const buffer = Buffer.alloc(size);
        const result = await handler.read(file, buffer, start);
        _assert.strict.deepEqual(result.buffer, buffer);
        _assert.strict.deepEqual(result, {
          buffer,
          bytesRead: Math.min(size, TEST_DATA_LEN - start)
        });
      });
    });
    (0, _test.describe)('#readFile', () => {
      (0, _test.it)('returns a buffer containing the contents of the file', async () => {
        await handler.outputFile('file', TEST_DATA);
        _assert.strict.deepEqual(await handler.readFile('file'), TEST_DATA);
      });
      (0, _test.it)('throws on missing file', async () => {
        await handler.unlink('file');
        const error = await rejectionOf(handler.readFile('file'));
        _assert.strict.equal(error.code, 'ENOENT');
      });
    });
    (0, _test.describe)('#rename()', () => {
      (0, _test.it)(`should rename the file`, async () => {
        await handler.outputFile('file', TEST_DATA);
        await handler.rename('file', `file2`);
        _assert.strict.deepEqual(await handler.list('.'), ['file2']);
        _assert.strict.deepEqual(await handler.readFile(`file2`), TEST_DATA);
      });
      (0, _test.it)(`should rename the file and create dest directory`, async () => {
        await handler.outputFile('file', TEST_DATA);
        await handler.rename('file', `sub/file2`);
        _assert.strict.deepEqual(await handler.list('sub'), ['file2']);
        _assert.strict.deepEqual(await handler.readFile(`sub/file2`), TEST_DATA);
      });
      (0, _test.it)(`should fail with enoent if source file is missing`, async () => {
        const error = await rejectionOf(handler.rename('file', `sub/file2`));
        _assert.strict.equal(error.code, 'ENOENT');
      });
    });
    (0, _test.describe)('#rmdir()', () => {
      (0, _test.it)('should remove an empty directory', async () => {
        await handler.mkdir('dir');
        await handler.rmdir('dir');
        _assert.strict.deepEqual(await handler.list('.'), []);
      });
      (0, _test.it)(`should throw on non-empty directory`, async () => {
        await handler.outputFile('dir/file', '');
        const error = await rejectionOf(handler.rmdir('.'));
        _assert.strict.equal(error.code, 'ENOTEMPTY');
        await handler.unlink('dir/file');
      });
      (0, _test.it)('does not throw on missing directory', async () => {
        await handler.rmdir('dir');
      });
    });
    (0, _test.describe)('#rmtree', () => {
      (0, _test.it)(`should remove a directory resursively`, async () => {
        await handler.outputFile('dir/file', '');
        await handler.rmtree('dir');
        _assert.strict.deepEqual(await handler.list('.'), []);
      });
    });
    (0, _test.describe)('#test()', () => {
      (0, _test.it)('tests the remote appears to be working', async () => {
        const answer = await handler.test();
        _assert.strict.equal(answer.success, true);
        _assert.strict.equal(typeof answer.writeRate, 'number');
        _assert.strict.equal(typeof answer.readRate, 'number');
      });
    });
    (0, _test.describe)('#unlink()', () => {
      (0, _test.it)(`should remove the file`, async () => {
        await handler.outputFile('file', TEST_DATA);
        await handler.unlink('file');
        _assert.strict.deepEqual(await handler.list('.'), []);
      });
      (0, _test.it)('does not throw on missing file', async () => {
        await handler.unlink('file');
      });
    });
    (0, _test.describe)('#write()', () => {
      (0, _test.beforeEach)(() => handler.outputFile('file', TEST_DATA));
      (0, _test.afterEach)(() => handler.unlink('file'));
      const PATCH_DATA_LEN = Math.ceil(TEST_DATA_LEN / 2);
      const PATCH_DATA = unsecureRandomBytes(PATCH_DATA_LEN);
      (0, _lodash.forOwn)({
        'dont increase file size': (() => {
          const offset = (0, _lodash.random)(0, TEST_DATA_LEN - PATCH_DATA_LEN);
          const expected = Buffer.from(TEST_DATA);
          PATCH_DATA.copy(expected, offset);
          return {
            offset,
            expected
          };
        })(),
        'increase file size': (() => {
          const offset = (0, _lodash.random)(TEST_DATA_LEN - PATCH_DATA_LEN + 1, TEST_DATA_LEN);
          const expected = Buffer.alloc(offset + PATCH_DATA_LEN);
          TEST_DATA.copy(expected);
          PATCH_DATA.copy(expected, offset);
          return {
            offset,
            expected
          };
        })()
      }, ({
        offset,
        expected
      }, title) => {
        (0, _test.describe)(title, () => {
          testWithFileDescriptor('file', 'r+', async ({
            file
          }) => {
            await handler.write(file, PATCH_DATA, offset);
            _assert.strict.deepEqual(await handler.readFile('file'), expected);
          });
        });
      });
    });
    (0, _test.describe)('#truncate()', () => {
      (0, _test.afterEach)(() => handler.unlink('file'));
      (0, _lodash.forOwn)({
        'shrinks file': (() => {
          const length = (0, _lodash.random)(0, TEST_DATA_LEN);
          const expected = TEST_DATA.slice(0, length);
          return {
            length,
            expected
          };
        })(),
        'grows file': (() => {
          const length = (0, _lodash.random)(TEST_DATA_LEN, TEST_DATA_LEN * 2);
          const expected = Buffer.alloc(length);
          TEST_DATA.copy(expected);
          return {
            length,
            expected
          };
        })()
      }, ({
        length,
        expected
      }, title) => {
        (0, _test.it)(title, async () => {
          await handler.outputFile('file', TEST_DATA);
          await handler.truncate('file', length);
          _assert.strict.deepEqual(await handler.readFile('file'), expected);
        });
      });
    });
  });
});
//# sourceMappingURL=fs.test.js.map