"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var _df = _interopRequireDefault(require("@sindresorhus/df"));
var _fsExtra = _interopRequireDefault(require("fs-extra"));
var _properLockfile = _interopRequireDefault(require("proper-lockfile"));
var _log = require("@xen-orchestra/log");
var _promiseToolbox = require("promise-toolbox");
var _abstract = _interopRequireDefault(require("./abstract"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const {
  info,
  warn
} = (0, _log.createLogger)('xo:fs:local');
async function addSyncStackTrace(fn, ...args) {
  const stackContainer = new Error();
  try {
    return await fn.apply(this, args);
  } catch (error) {
    let {
      stack
    } = stackContainer;
    stack = stack.slice(stack.indexOf('\n') + 1);
    error.stack = [error.stack, 'From:', stack].join('\n');
    throw error;
  }
}
function dontAddSyncStackTrace(fn, ...args) {
  return fn.apply(this, args);
}
class LocalHandler extends _abstract.default {
  constructor(remote, opts = {}) {
    super(remote);
    this._addSyncStackTrace = opts.syncStackTraces ?? true ? addSyncStackTrace : dontAddSyncStackTrace;
    this._retriesOnEagain = {
      delay: 1e3,
      retries: 9,
      ...opts.retriesOnEagain,
      when: {
        code: 'EAGAIN'
      }
    };
  }
  get type() {
    return 'file';
  }
  _getRealPath() {
    return this._remote.path;
  }
  _getFilePath(file) {
    return this._getRealPath() + file;
  }
  async _closeFile(fd) {
    return this._addSyncStackTrace(_fsExtra.default.close, fd);
  }
  async _copy(oldPath, newPath) {
    return this._addSyncStackTrace(_fsExtra.default.copy, this._getFilePath(oldPath), this._getFilePath(newPath));
  }
  async _createReadStream(file, options) {
    if (typeof file === 'string') {
      const stream = _fsExtra.default.createReadStream(this._getFilePath(file), options);
      await this._addSyncStackTrace(_promiseToolbox.fromEvent, stream, 'open');
      return stream;
    }
    return _fsExtra.default.createReadStream('', {
      autoClose: false,
      ...options,
      fd: file.fd
    });
  }
  async _createWriteStream(file, options) {
    if (typeof file === 'string') {
      const stream = _fsExtra.default.createWriteStream(this._getFilePath(file), options);
      await this._addSyncStackTrace(_promiseToolbox.fromEvent, stream, 'open');
      return stream;
    }
    return _fsExtra.default.createWriteStream('', {
      autoClose: false,
      ...options,
      fd: file.fd
    });
  }
  async _getInfo() {
    const info = await _df.default.file(this._getFilePath('/'));
    Object.keys(info).forEach(key => {
      if (Number.isNaN(info[key])) {
        delete info[key];
      }
    });
    return info;
  }
  async _getSize(file) {
    const stats = await this._addSyncStackTrace(_fsExtra.default.stat, this._getFilePath(typeof file === 'string' ? file : file.path));
    return stats.size;
  }
  async _list(dir) {
    return this._addSyncStackTrace(_fsExtra.default.readdir, this._getFilePath(dir));
  }
  async _lock(path) {
    const acquire = _properLockfile.default.lock.bind(undefined, this._getFilePath(path), {
      async onCompromised(error) {
        warn('lock compromised', {
          error
        });
        try {
          release = await acquire();
          info('compromised lock was reacquired');
        } catch (error) {
          warn('compromised lock could not be reacquired', {
            error
          });
        }
      }
    });
    let release = await this._addSyncStackTrace(acquire);
    return async () => {
      try {
        await this._addSyncStackTrace(release);
      } catch (error) {
        warn('lock could not be released', {
          error
        });
      }
    };
  }
  _mkdir(dir, {
    mode
  }) {
    return this._addSyncStackTrace(_fsExtra.default.mkdir, this._getFilePath(dir), {
      mode
    });
  }
  async _openFile(path, flags) {
    return this._addSyncStackTrace(_fsExtra.default.open, this._getFilePath(path), flags);
  }
  async _read(file, buffer, position) {
    const needsClose = typeof file === 'string';
    file = needsClose ? await this._addSyncStackTrace(_fsExtra.default.open, this._getFilePath(file), 'r') : file.fd;
    try {
      return await this._addSyncStackTrace(_fsExtra.default.read, file, buffer, 0, buffer.length, position === undefined ? null : position);
    } finally {
      if (needsClose) {
        await this._addSyncStackTrace(_fsExtra.default.close, file);
      }
    }
  }
  async _readFile(file, options) {
    const filePath = this._getFilePath(file);
    return await this._addSyncStackTrace(_promiseToolbox.retry, () => _fsExtra.default.readFile(filePath, options), this._retriesOnEagain);
  }
  async _rename(oldPath, newPath) {
    return this._addSyncStackTrace(_fsExtra.default.rename, this._getFilePath(oldPath), this._getFilePath(newPath));
  }
  async _rmdir(dir) {
    return this._addSyncStackTrace(_fsExtra.default.rmdir, this._getFilePath(dir));
  }
  async _sync() {
    const path = this._getRealPath('/');
    await this._addSyncStackTrace(_fsExtra.default.ensureDir, path);
    await this._addSyncStackTrace(_fsExtra.default.access, path, _fsExtra.default.R_OK | _fsExtra.default.W_OK);
  }
  _truncate(file, len) {
    return this._addSyncStackTrace(_fsExtra.default.truncate, this._getFilePath(file), len);
  }
  async _unlink(file) {
    const filePath = this._getFilePath(file);
    return await this._addSyncStackTrace(_promiseToolbox.retry, () => _fsExtra.default.unlink(filePath), this._retriesOnEagain);
  }
  _writeFd(file, buffer, position) {
    return this._addSyncStackTrace(_fsExtra.default.write, file.fd, buffer, 0, buffer.length, position);
  }
  _writeFile(file, data, {
    flags
  }) {
    return this._addSyncStackTrace(_fsExtra.default.writeFile, this._getFilePath(file), data, {
      flag: flags
    });
  }
}
exports.default = LocalHandler;
//# sourceMappingURL=local.js.map