"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ThrottleGroup = void 0;
const limiter_1 = require("limiter");
const throttle_1 = require("./throttle");
class ThrottleGroup {
    constructor(options) {
        var _a;
        if (options.rate <= 0) {
            throw new Error('Rate must be greater than 0');
        }
        if (options.burst && options.burst < options.rate) {
            throw new Error('Burst must be greater than or equal to the rate');
        }
        if (options.chunkSize !== undefined && options.chunkSize <= 0) {
            throw new Error('Chunk size must be greater than 0');
        }
        if (options.chunkSize && options.chunkSize > options.rate) {
            throw new Error('Chunk size must be less than or equal to the rate');
        }
        this._rate = options.rate;
        this._burst = options.burst || this.rate;
        this._chunkSize = (_a = options.chunkSize) !== null && _a !== void 0 ? _a : this.calculateChunkSize(this.rate);
        this._bucket = new limiter_1.TokenBucket({
            tokensPerInterval: this.rate,
            bucketSize: this.burst,
            interval: 'second',
        });
    }
    get rate() {
        return this._rate;
    }
    get burst() {
        return this._burst;
    }
    get bucket() {
        return this._bucket;
    }
    get chunkSize() {
        return this._chunkSize;
    }
    /**
     * Update the sustained rate of the throttle.
     * Calling this method will recalculate the chunk size.
     * In case the the burst rate is less than the sustained rate,
     * the burst rate will be set to the updated sustained rate.
     * @param rate The rate of the throttle in bytes per second
     */
    setRate(rate) {
        rate = Math.floor(rate);
        if (this.bucket.bucketSize < rate) {
            this.bucket.bucketSize = rate;
            this._burst = rate;
        }
        this.bucket.tokensPerInterval = rate;
        this._rate = rate;
        this._chunkSize = this.calculateChunkSize(rate);
    }
    /**
     * Update the burst rate (also known as bucket size) of the throttle.
     * @param burst The burst value of the throttle in bytes per second
     * @throws {Error} If the burst value is less than the rate
     */
    setBurst(burst) {
        burst = Math.floor(burst);
        if (burst < this.bucket.tokensPerInterval) {
            throw new Error('Burst must be greater than or equal to the rate');
        }
        this.bucket.bucketSize = burst;
        this._burst = burst;
    }
    /**
     * Update the processing chunk size of the throttle.
     * @param chunkSize The chunk size in bytes that received chunks will be split into
     */
    setChunkSize(chunkSize) {
        chunkSize = Math.floor(chunkSize);
        if (chunkSize > this.rate) {
            throw new Error('Chunk size must be less than or equal to the rate');
        }
        this._chunkSize = chunkSize;
    }
    createThrottle() {
        return new throttle_1.Throttle(this);
    }
    calculateChunkSize(chunkSize) {
        return Math.floor(chunkSize / 10);
    }
}
exports.ThrottleGroup = ThrottleGroup;
//# sourceMappingURL=group.js.map