"use strict";

exports.__esModule = true;
exports.limitConcurrency = exports.limitMethod = exports.BYPASS_QUEUE = exports.FAIL_ON_QUEUE = void 0;

function Deferred(fn, thisArg, args, resolve, reject) {
  this.args = args;
  this.fn = fn;
  this.reject = reject;
  this.resolve = resolve;
  this.thisArg = thisArg;
}

function next() {
  var d = this.pop();

  if (d === undefined) {
    ++this.concurrency;
  } else {
    try {
      d.resolve(d.fn.apply(d.thisArg, d.args));
    } catch (error) {
      d.reject(error);
    }
  }
}

function Queue(concurrency) {
  this.concurrency = concurrency;
  this.next = next.bind(this);
  this._s1 = [];
  this._s2 = [];
}

Queue.prototype.push = function (value) {
  this._s1.push(value);
};

Queue.prototype.pop = function () {
  var s2 = this._s2;

  if (s2.length === 0) {
    var s1 = this._s1;

    if (s1.length === 0) {
      return;
    }

    this._s1 = s2;
    s2 = this._s2 = s1.reverse();
  }

  return s2.pop();
};

var getSymbol = typeof Symbol === "function" ? function (key) {
  return Symbol.for("limit-concurrency-decorator/" + key);
} : function (key) {
  return "@@limit-concurrency-decorator/" + key;
};
var FAIL_ON_QUEUE = getSymbol("FAIL_ON_QUEUE");
exports.FAIL_ON_QUEUE = FAIL_ON_QUEUE;
var BYPASS_QUEUE = getSymbol("BYPASS_QUEUE");
exports.BYPASS_QUEUE = BYPASS_QUEUE;

var defaultTermination = function defaultTermination(promise) {
  return promise;
};

var slice = Array.prototype.slice;

var makeLimiter = function makeLimiter(getQueue, termination) {
  if (termination === void 0) {
    termination = defaultTermination;
  }

  return function (fn) {
    return function () {
      var _this = this;

      var queue = getQueue(this);
      var canRun = queue.concurrency > 0;
      var argStart = 0;
      var length = arguments.length;

      if (argStart < length) {
        var maybeFlag = arguments[argStart++];

        if (maybeFlag === BYPASS_QUEUE) {
          canRun = true;
        } else if (maybeFlag === FAIL_ON_QUEUE) {
          if (!canRun) {
            return Promise.reject(new Error("no available place in queue"));
          }
        } else {
          --argStart;
        }
      }

      var args = slice.call(arguments, argStart);
      var promise;

      if (canRun) {
        --queue.concurrency;

        try {
          promise = fn.apply(this, args);

          if (promise == null || typeof promise.then !== "function") {
            promise = Promise.resolve(promise);
          }
        } catch (error) {
          promise = Promise.reject(promise);
        }
      } else {
        promise = new Promise(function (resolve, reject) {
          return queue.push(new Deferred(fn, _this, args, resolve, reject));
        });
      }

      var next = queue.next;
      termination(promise).then(next, next);
      return promise;
    };
  };
};

var limitFunction = function limitFunction(concurrency, opts) {
  var queue = new Queue(concurrency);
  return makeLimiter(function () {
    return queue;
  }, opts);
};

var limitMethod = function limitMethod(concurrency, opts) {
  var queues = new WeakMap();
  return makeLimiter(function (obj) {
    var queue = queues.get(obj);

    if (queue === undefined) {
      queue = new Queue(concurrency);
      queues.set(obj, queue);
    }

    return queue;
  }, opts);
};

exports.limitMethod = limitMethod;

var limitConcurrency = function limitConcurrency() {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  var method = false;
  var wrap;
  return function (target, key, descriptor) {
    if (key === undefined) {
      if (wrap === undefined) {
        wrap = limitFunction.apply(void 0, args);
      } else if (method) {
        throw new Error("the same decorator cannot be used between function and method");
      }

      return wrap(target);
    }

    if (wrap === undefined) {
      method = true;
      wrap = limitMethod.apply(void 0, args);
    } else if (!method) {
      throw new Error("the same decorator cannot be used between function and method");
    }

    if ("value" in descriptor) {
      descriptor.value = wrap(descriptor.value);
    } else {
      var get = descriptor.get;

      descriptor.get = function () {
        return wrap(get.call(this));
      };
    }

    return descriptor;
  };
};

exports.limitConcurrency = limitConcurrency;
//# sourceMappingURL=index.js.map