"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.debug = void 0;
var _lodash = require("lodash");
var _plan = _interopRequireDefault(require("./plan"));
var _utils = require("./utils");
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }
function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }
const debug = str => (0, _utils.debug)(`density: ${str}`);
exports.debug = debug;
class DensityPlan extends _plan.default {
  _checkResourcesThresholds(objects, averages) {
    const {
      low
    } = this._thresholds.memoryFree;
    return (0, _lodash.filter)(objects, object => {
      const {
        memory,
        memoryFree = memory
      } = averages[object.id];
      return memoryFree > low;
    });
  }
  execute() {
    var _this = this;
    return _asyncToGenerator(function* () {
      yield _this._processAntiAffinity();
      const hosts = _this._getHosts();
      const results = yield _this._getHostStatsAverages({
        hosts,
        toOptimizeOnly: true
      });
      if (!results) {
        return;
      }
      const {
        toOptimize
      } = results;
      let {
        averages: hostsAverages
      } = results;
      const pools = yield _this._getPlanPools();
      let optimizationsCount = 0;
      for (const hostToOptimize of toOptimize) {
        const {
          id: hostId,
          $poolId: poolId
        } = hostToOptimize;
        const {
          master: masterId
        } = pools[poolId];
        if (masterId === hostId) {
          continue;
        }
        if (hostToOptimize.powerOnMode === '') {
          debug(`Host (${hostId}) does not have a power on mode.`);
          continue;
        }
        let poolMaster;
        const poolHosts = [];
        const masters = [];
        const otherHosts = [];
        for (const dest of hosts) {
          const {
            id: destId,
            $poolId: destPoolId
          } = dest;
          if (destId === hostId) {
            continue;
          }
          if (destPoolId === poolId) {
            if (destId === masterId) {
              poolMaster = dest;
            } else {
              poolHosts.push(dest);
            }
          } else if (destId === pools[destPoolId].master) {
            masters.push(dest);
          } else {
            otherHosts.push(dest);
          }
        }
        const simulResults = yield _this._simulate({
          host: hostToOptimize,
          destinations: [[poolMaster], poolHosts, masters, otherHosts],
          hostsAverages: (0, _lodash.clone)(hostsAverages)
        });
        if (simulResults) {
          hostsAverages = simulResults.hostsAverages;
          yield _this._migrate(hostToOptimize, simulResults.moves);
          optimizationsCount++;
        }
      }
      debug(`Density mode: ${optimizationsCount} optimizations.`);
    })();
  }
  _simulate({
    host,
    destinations,
    hostsAverages
  }) {
    var _this2 = this;
    return _asyncToGenerator(function* () {
      const {
        id: hostId
      } = host;
      debug(`Try to optimize Host (${hostId}).`);
      const vms = (0, _lodash.filter)(_this2._getAllRunningVms(), vm => vm.$container === hostId);
      const vmsAverages = yield _this2._getVmsAverages(vms, {
        [host.id]: host
      });
      for (const vm of vms) {
        if (!vm.xenTools) {
          debug(`VM (${vm.id}) of Host (${hostId}) does not support pool migration.`);
          return;
        }
        for (const tag of vm.tags) {
          if (_this2._antiAffinityTags.includes(tag)) {
            debug(`VM (${vm.id}) of Host (${hostId}) cannot be migrated. It contains anti-affinity tag '${tag}'.`);
            return;
          }
        }
      }
      vms.sort((a, b) => vmsAverages[b.id].memory - vmsAverages[a.id].memory);
      const simulResults = {
        hostsAverages,
        moves: []
      };
      for (const vm of vms) {
        let move;
        for (const subDestinations of destinations) {
          move = _this2._testMigration({
            vm,
            destinations: subDestinations,
            hostsAverages,
            vmsAverages
          });
          if (move) {
            simulResults.moves.push(move);
            break;
          }
        }
        if (!move) {
          return;
        }
      }
      return simulResults;
    })();
  }
  _testMigration({
    vm,
    destinations,
    hostsAverages,
    vmsAverages
  }) {
    const {
      _thresholds: {
        critical: criticalThreshold
      }
    } = this;
    destinations.sort((a, b) => hostsAverages[a.id].memoryFree - hostsAverages[b.id].memoryFree);
    for (const destination of destinations) {
      const destinationAverages = hostsAverages[destination.id];
      const vmAverages = vmsAverages[vm.id];
      if (destinationAverages.cpu + vmAverages.cpu >= criticalThreshold || destinationAverages.memoryFree - vmAverages.memory <= criticalThreshold) {
        continue;
      }
      destinationAverages.cpu += vmAverages.cpu;
      destinationAverages.memoryFree -= vmAverages.memory;
      return {
        vm,
        destination
      };
    }
  }
  _migrate(srcHost, moves) {
    var _this3 = this;
    return _asyncToGenerator(function* () {
      const xapiSrc = _this3.xo.getXapi(srcHost.id);
      const fmtSrcHost = `${srcHost.id} "${srcHost.name_label}"`;
      yield Promise.all((0, _lodash.map)(moves, move => {
        const {
          vm,
          destination
        } = move;
        const xapiDest = _this3.xo.getXapi(destination);
        debug(`Migrate VM (${vm.id} "${vm.name_label}") to Host (${destination.id} "${destination.name_label}") from Host (${fmtSrcHost}).`);
        return xapiDest.migrateVm(vm._xapiId, _this3.xo.getXapi(destination), destination._xapiId);
      }));
      debug(`Shutdown Host (${fmtSrcHost}).`);
      try {
        yield xapiSrc.shutdownHost(srcHost.id);
      } catch (error) {
        debug(`Unable to shutdown Host (${fmtSrcHost}).`, {
          error
        });
      }
    })();
  }
}
exports.default = DensityPlan;
//# sourceMappingURL=density-plan.js.map