"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.configurationSchema = void 0;
var _cron = require("@xen-orchestra/cron");
var _lodash = require("lodash");
var _densityPlan = _interopRequireDefault(require("./density-plan"));
var _performancePlan = _interopRequireDefault(require("./performance-plan"));
var _simplePlan = _interopRequireDefault(require("./simple-plan"));
var _plan = require("./plan");
var _utils = require("./utils");
const _excluded = ["name", "pools"];
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
function _objectWithoutProperties(source, excluded) { if (source == null) return {}; var target = _objectWithoutPropertiesLoose(source, excluded); var key, i; if (Object.getOwnPropertySymbols) { var sourceSymbolKeys = Object.getOwnPropertySymbols(source); for (i = 0; i < sourceSymbolKeys.length; i++) { key = sourceSymbolKeys[i]; if (excluded.indexOf(key) >= 0) continue; if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue; target[key] = source[key]; } } return target; }
function _objectWithoutPropertiesLoose(source, excluded) { if (source == null) return {}; var target = {}; var sourceKeys = Object.keys(source); var key, i; for (i = 0; i < sourceKeys.length; i++) { key = sourceKeys[i]; if (excluded.indexOf(key) >= 0) continue; target[key] = source[key]; } return target; }
function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }
function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }
const PERFORMANCE_MODE = 0;
const DENSITY_MODE = 1;
const configurationSchema = {
  type: 'object',
  properties: {
    plans: {
      type: 'array',
      description: 'an array of plans',
      title: 'Plans',
      items: {
        type: 'object',
        title: 'Plan',
        properties: {
          name: {
            type: 'string',
            title: 'Name'
          },
          mode: {
            enum: ['Performance mode', 'Density mode', 'Simple mode'],
            title: 'Mode'
          },
          pools: {
            type: 'array',
            description: 'list of pools where to apply the policy',
            items: {
              type: 'string',
              $type: 'Pool'
            }
          },
          thresholds: {
            type: 'object',
            title: 'Critical thresholds',
            properties: {
              cpu: {
                type: 'integer',
                title: 'CPU (%)',
                default: _plan.DEFAULT_CRITICAL_THRESHOLD_CPU
              },
              memoryFree: {
                type: 'integer',
                title: 'RAM, Free memory (MB)',
                default: _plan.DEFAULT_CRITICAL_THRESHOLD_MEMORY_FREE
              }
            }
          },
          excludedHosts: {
            type: 'array',
            title: 'Excluded hosts',
            description: 'list of hosts that are not affected by the plan',
            items: {
              type: 'string',
              $type: 'Host'
            }
          },
          antiAffinityTags: {
            type: 'array',
            title: 'Anti-affinity tags',
            description: 'list of VM tags to force place VMs on different hosts',
            items: {
              type: 'string',
              $type: 'Tag'
            }
          }
        },
        required: ['name', 'mode', 'pools']
      },
      minItems: 1
    },
    ignoredVmTags: {
      type: 'array',
      title: 'Ignored VM tags',
      description: 'list of VM tags to never migrate specific VMs',
      items: {
        type: 'string',
        $type: 'Tag'
      }
    }
  },
  additionalProperties: false
};
exports.configurationSchema = configurationSchema;
class LoadBalancerPlugin {
  constructor(xo) {
    var _this = this;
    this.xo = xo;
    this._job = (0, _cron.createSchedule)(`*/${_utils.EXECUTION_DELAY} * * * *`).createJob(_asyncToGenerator(function* () {
      try {
        yield _this._executePlans();
      } catch (error) {
        console.error('[WARN] scheduled function:', error && error.stack || error);
      }
    }));
  }
  configure({
    plans,
    ignoredVmTags = []
  }) {
    var _this2 = this;
    return _asyncToGenerator(function* () {
      _this2._plans = [];
      _this2._poolIds = [];
      _this2._globalOptions = {
        ignoredVmTags: new Set(ignoredVmTags)
      };
      if (plans) {
        for (const plan of plans) {
          _this2._addPlan(plan.mode === 'Performance mode' ? PERFORMANCE_MODE : DENSITY_MODE, plan);
        }
      }
    })();
  }
  load() {
    this._job.start();
  }
  unload() {
    this._job.stop();
  }
  _addPlan(mode, _ref2) {
    let {
        name,
        pools
      } = _ref2,
      options = _objectWithoutProperties(_ref2, _excluded);
    pools = (0, _lodash.uniq)(pools);
    if ((0, _lodash.intersection)(pools, this._poolIds).length > 0) {
      throw new Error(`Pool(s) already included in an other plan: ${pools}`);
    }
    this._poolIds = this._poolIds.concat(pools);
    let plan;
    if (mode === PERFORMANCE_MODE) {
      plan = new _performancePlan.default(this.xo, name, pools, options, this._globalOptions);
    } else if (mode === DENSITY_MODE) {
      plan = new _densityPlan.default(this.xo, name, pools, options, this._globalOptions);
    } else {
      plan = new _simplePlan.default(this.xo, name, pools, options, this._globalOptions);
    }
    this._plans.push(plan);
  }
  _executePlans() {
    (0, _utils.debug)('Execute plans!');
    return Promise.all(this._plans.map(plan => plan.execute()));
  }
}
var _default = ({
  xo
}) => new LoadBalancerPlugin(xo);
exports.default = _default;
//# sourceMappingURL=index.js.map