"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.debugAffinity = exports.DEFAULT_CRITICAL_THRESHOLD_MEMORY_FREE = exports.DEFAULT_CRITICAL_THRESHOLD_CPU = void 0;
var _lodash = require("lodash");
var _util = require("util");
var _utils = require("./utils");
function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }
function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }
const MINUTES_OF_HISTORICAL_DATA = 30;
const DEFAULT_CRITICAL_THRESHOLD_CPU = 90.0;
exports.DEFAULT_CRITICAL_THRESHOLD_CPU = DEFAULT_CRITICAL_THRESHOLD_CPU;
const DEFAULT_CRITICAL_THRESHOLD_MEMORY_FREE = 1000.0;
exports.DEFAULT_CRITICAL_THRESHOLD_MEMORY_FREE = DEFAULT_CRITICAL_THRESHOLD_MEMORY_FREE;
const HIGH_THRESHOLD_FACTOR = 0.85;
const LOW_THRESHOLD_FACTOR = 0.65;
const HIGH_THRESHOLD_MEMORY_FREE_FACTOR = 1.2;
const LOW_THRESHOLD_MEMORY_FREE_FACTOR = 1.5;
const numberOrDefault = (value, def) => value >= 0 ? value : def;
const debugAffinity = str => (0, _utils.debug)(`anti-affinity: ${str}`);
exports.debugAffinity = debugAffinity;
function computeAverage(values, nPoints) {
  if (values === undefined) {
    return;
  }
  let sum = 0;
  let tot = 0;
  const {
    length
  } = values;
  const start = nPoints !== undefined ? length - nPoints : 0;
  for (let i = start; i < length; i++) {
    const value = values[i];
    sum += value || 0;
    if (value) {
      tot += 1;
    }
  }
  return sum / tot;
}
function computeResourcesAverage(objects, objectsStats, nPoints) {
  const averages = {};
  for (const object of objects) {
    const {
      id
    } = object;
    const {
      stats
    } = objectsStats[id];
    averages[id] = {
      cpu: computeAverage((0, _lodash.map)(stats.cpus, cpu => computeAverage(cpu, nPoints))),
      nCpus: (0, _lodash.size)(stats.cpus),
      memoryFree: computeAverage(stats.memoryFree, nPoints),
      memory: computeAverage(stats.memory, nPoints)
    };
  }
  return averages;
}
function computeResourcesAverageWithWeight(averages1, averages2, ratio) {
  const averages = {};
  for (const id in averages1) {
    const objectAverages = averages[id] = {};
    for (const averageName in averages1[id]) {
      const average1 = averages1[id][averageName];
      if (average1 === undefined) {
        continue;
      }
      objectAverages[averageName] = average1 * ratio + averages2[id][averageName] * (1 - ratio);
    }
  }
  return averages;
}
function setRealCpuAverageOfVms(vms, vmsAverages, nCpus) {
  for (const vm of vms) {
    const averages = vmsAverages[vm.id];
    averages.cpu *= averages.nCpus / nCpus;
  }
}
class Plan {
  constructor(xo, name, poolIds, {
    excludedHosts,
    thresholds,
    antiAffinityTags = []
  }, globalOptions) {
    this.xo = xo;
    this._name = name;
    this._poolIds = poolIds;
    this._excludedHosts = excludedHosts;
    this._thresholds = {
      cpu: {
        critical: numberOrDefault(thresholds && thresholds.cpu, DEFAULT_CRITICAL_THRESHOLD_CPU)
      },
      memoryFree: {
        critical: numberOrDefault(thresholds && thresholds.memoryFree, DEFAULT_CRITICAL_THRESHOLD_MEMORY_FREE) * 1024 * 1024
      }
    };
    this._antiAffinityTags = antiAffinityTags;
    this._globalOptions = globalOptions;
    for (const key in this._thresholds) {
      const attr = this._thresholds[key];
      const {
        critical
      } = attr;
      if (key === 'memoryFree') {
        attr.high = critical * HIGH_THRESHOLD_MEMORY_FREE_FACTOR;
        attr.low = critical * LOW_THRESHOLD_MEMORY_FREE_FACTOR;
      } else {
        attr.high = critical * HIGH_THRESHOLD_FACTOR;
        attr.low = critical * LOW_THRESHOLD_FACTOR;
      }
    }
  }
  execute() {
    throw new Error('Not implemented');
  }
  _getHostStatsAverages({
    hosts,
    toOptimizeOnly = false
  }) {
    var _this = this;
    return _asyncToGenerator(function* () {
      const hostsStats = yield _this._getHostsStats(hosts, 'minutes');
      const avgNow = computeResourcesAverage(hosts, hostsStats, _utils.EXECUTION_DELAY);
      let toOptimize;
      if (toOptimizeOnly) {
        toOptimize = _this._checkResourcesThresholds(hosts, avgNow);
        if (toOptimize.length === 0) {
          (0, _utils.debug)('No hosts to optimize.');
          return;
        }
      }
      const avgBefore = computeResourcesAverage(hosts, hostsStats, MINUTES_OF_HISTORICAL_DATA);
      const avgWithRatio = computeResourcesAverageWithWeight(avgNow, avgBefore, 0.75);
      if (toOptimizeOnly) {
        toOptimize = _this._checkResourcesThresholds(toOptimize, avgWithRatio);
        if (toOptimize.length === 0) {
          (0, _utils.debug)('No hosts to optimize.');
          return;
        }
      }
      return {
        toOptimize,
        averages: avgWithRatio
      };
    })();
  }
  _checkResourcesThresholds() {
    throw new Error('Not implemented');
  }
  _getPlanPools() {
    const pools = {};
    try {
      for (const poolId of this._poolIds) {
        pools[poolId] = this.xo.getObject(poolId);
      }
    } catch (_) {
      return {};
    }
    return pools;
  }
  _getHosts({
    powerState = 'Running'
  } = {}) {
    return (0, _lodash.filter)(this.xo.getObjects(), object => object.type === 'host' && (0, _lodash.includes)(this._poolIds, object.$poolId) && object.power_state === powerState && !(0, _lodash.includes)(this._excludedHosts, object.id));
  }
  _getAllRunningVms() {
    return (0, _lodash.filter)(this.xo.getObjects(), object => object.type === 'VM' && object.power_state === 'Running' && !object.tags.some(tag => this._globalOptions.ignoredVmTags.has(tag)));
  }
  _getHostsStats(hosts, granularity) {
    var _this2 = this;
    return _asyncToGenerator(function* () {
      const hostsStats = {};
      yield Promise.all(hosts.map(host => _this2.xo.getXapiHostStats(host, granularity).then(hostStats => {
        hostsStats[host.id] = {
          nPoints: hostStats.stats.cpus[0].length,
          stats: hostStats.stats,
          averages: {}
        };
      })));
      return hostsStats;
    })();
  }
  _getVmsStats(vms, granularity) {
    var _this3 = this;
    return _asyncToGenerator(function* () {
      const vmsStats = {};
      yield Promise.all(vms.map(vm => _this3.xo.getXapiVmStats(vm, granularity).then(vmStats => {
        vmsStats[vm.id] = {
          nPoints: vmStats.stats.cpus[0].length,
          stats: vmStats.stats,
          averages: {}
        };
      })));
      return vmsStats;
    })();
  }
  _getVmsAverages(vms, hosts) {
    var _this4 = this;
    return _asyncToGenerator(function* () {
      const vmsStats = yield _this4._getVmsStats(vms, 'minutes');
      const vmsAverages = computeResourcesAverageWithWeight(computeResourcesAverage(vms, vmsStats, _utils.EXECUTION_DELAY), computeResourcesAverage(vms, vmsStats, MINUTES_OF_HISTORICAL_DATA), 0.75);
      for (const [hostId, hostVms] of Object.entries((0, _lodash.groupBy)(vms, '$container'))) {
        setRealCpuAverageOfVms(hostVms, vmsAverages, hosts[hostId].CPUs.cpu_count);
      }
      return vmsAverages;
    })();
  }
  _processAntiAffinity() {
    var _this5 = this;
    return _asyncToGenerator(function* () {
      if (!_this5._antiAffinityTags.length) {
        return;
      }
      const allHosts = yield _this5._getHosts();
      if (allHosts.length <= 1) {
        return;
      }
      const idToHost = (0, _lodash.keyBy)(allHosts, 'id');
      const allVms = (0, _lodash.filter)(_this5._getAllRunningVms(), vm => vm.$container in idToHost);
      const taggedHosts = _this5._getAntiAffinityTaggedHosts(allHosts, allVms);
      const tagsDiff = {};
      for (const watchedTag of _this5._antiAffinityTags) {
        const getCount = fn => fn(taggedHosts.hosts, host => host.tags[watchedTag]).tags[watchedTag];
        const diff = getCount(_lodash.maxBy) - getCount(_lodash.minBy);
        if (diff > 1) {
          tagsDiff[watchedTag] = diff - 1;
        }
      }
      if ((0, _lodash.isEmpty)(tagsDiff)) {
        return;
      }
      debugAffinity('Try to apply anti-affinity policy.');
      debugAffinity(`VM tag count per host: ${(0, _util.inspect)(taggedHosts, {
        depth: null
      })}.`);
      debugAffinity(`Tags diff: ${(0, _util.inspect)(tagsDiff, {
        depth: null
      })}.`);
      const vmsAverages = yield _this5._getVmsAverages(allVms, idToHost);
      const {
        averages: hostsAverages
      } = yield _this5._getHostStatsAverages({
        hosts: allHosts
      });
      debugAffinity(`Hosts averages: ${(0, _util.inspect)(hostsAverages, {
        depth: null
      })}.`);
      const promises = [];
      for (const tag in tagsDiff) {
        promises.push(..._this5._processAntiAffinityTag({
          tag,
          vmsAverages,
          hostsAverages,
          taggedHosts,
          idToHost
        }));
      }
      debugAffinity(`VM tag count per host after migration: ${(0, _util.inspect)(taggedHosts, {
        depth: null
      })}.`);
      return Promise.all(promises);
    })();
  }
  _processAntiAffinityTag({
    tag,
    vmsAverages,
    hostsAverages,
    taggedHosts,
    idToHost
  }) {
    const promises = [];
    while (true) {
      const sources = (0, _lodash.sortBy)((0, _lodash.filter)(taggedHosts.hosts, host => host.tags[tag] > 1), [host => host.tags[tag], host => -hostsAverages[host.id].memoryFree]);
      for (let sourceIndex = sources.length; sourceIndex >= 0; --sourceIndex) {
        if (sourceIndex === 0) {
          return promises;
        }
        const sourceHost = sources[sourceIndex - 1];
        const destinations = (0, _lodash.sortBy)((0, _lodash.filter)(taggedHosts.hosts, host => host.id !== sourceHost.id && host.tags[tag] + 1 < sourceHost.tags[tag]), [host => host.tags[tag], host => host.poolId !== sourceHost.poolId, host => -hostsAverages[host.id].memoryFree]);
        if (!destinations.length) {
          return promises;
        }
        const sourceVms = (0, _lodash.filter)(sourceHost.vms, vm => vm.tags.includes(tag));
        let destinationHost;
        let vm;
        for (const destination of destinations) {
          destinationHost = destination;
          debugAffinity(`Host candidate: ${sourceHost.id} -> ${destinationHost.id}.`);
          const vms = (0, _lodash.filter)(sourceVms, vm => hostsAverages[destinationHost.id].memoryFree >= vmsAverages[vm.id].memory);
          debugAffinity(`Tagged VM ("${tag}") candidates to migrate from host ${sourceHost.id}: ${(0, _util.inspect)((0, _lodash.map)(vms, 'id'))}.`);
          vm = this._getAntiAffinityVmToMigrate({
            vms,
            vmsAverages,
            hostsAverages,
            taggedHosts,
            sourceHost,
            destinationHost
          });
          if (vm) {
            break;
          }
        }
        if (!vm) {
          continue;
        }
        const source = idToHost[sourceHost.id];
        const destination = idToHost[destinationHost.id];
        debugAffinity(`Migrate VM (${vm.id} "${vm.name_label}") to Host (${destinationHost.id} "${destination.name_label}") from Host (${sourceHost.id} "${source.name_label}").`);
        for (const tag of vm.tags) {
          if (this._antiAffinityTags.includes(tag)) {
            sourceHost.tags[tag]--;
            destinationHost.tags[tag]++;
          }
        }
        const destinationAverages = hostsAverages[destinationHost.id];
        const vmAverages = vmsAverages[vm.id];
        destinationAverages.cpu += vmAverages.cpu;
        destinationAverages.memoryFree -= vmAverages.memory;
        delete sourceHost.vms[vm.id];
        promises.push(this.xo.getXapi(source).migrateVm(vm._xapiId, this.xo.getXapi(destination), destination._xapiId));
        break;
      }
    }
  }
  _getAntiAffinityTaggedHosts(hosts, vms) {
    const tagCount = {};
    for (const tag of this._antiAffinityTags) {
      tagCount[tag] = 0;
    }
    const taggedHosts = {};
    for (const host of hosts) {
      const tags = {};
      for (const tag of this._antiAffinityTags) {
        tags[tag] = 0;
      }
      const taggedHost = taggedHosts[host.id] = {
        id: host.id,
        poolId: host.$poolId,
        tags,
        vms: {}
      };
      Object.defineProperties(taggedHost, {
        poolId: {
          enumerable: false
        },
        vms: {
          enumerable: false
        }
      });
    }
    for (const vm of vms) {
      const hostId = vm.$container;
      if (!(hostId in taggedHosts)) {
        continue;
      }
      const taggedHost = taggedHosts[hostId];
      for (const tag of vm.tags) {
        if (this._antiAffinityTags.includes(tag)) {
          tagCount[tag]++;
          taggedHost.tags[tag]++;
          taggedHost.vms[vm.id] = vm;
        }
      }
    }
    return {
      tagCount,
      hosts: Object.values(taggedHosts)
    };
  }
  _computeAntiAffinityVariance(taggedHosts) {
    let variance = 0;
    const {
      hosts
    } = taggedHosts;
    for (const tag in taggedHosts.tagCount) {
      const k = hosts[0].tags[tag];
      let ex = 0;
      let ex2 = 0;
      for (const host of hosts) {
        const x = host.tags[tag];
        const diff = x - k;
        ex += diff;
        ex2 += diff * diff;
      }
      const n = hosts.length;
      variance += (ex2 - ex * ex / n) / n;
    }
    return variance;
  }
  _getAntiAffinityVmToMigrate({
    vms,
    vmsAverages,
    hostsAverages,
    taggedHosts,
    sourceHost,
    destinationHost
  }) {
    let bestVariance = this._computeAntiAffinityVariance(taggedHosts);
    let bestVm;
    for (const vm of vms) {
      const vmTags = (0, _lodash.filter)(vm.tags, tag => this._antiAffinityTags.includes(tag));
      for (const tag of vmTags) {
        sourceHost.tags[tag]--;
        destinationHost.tags[tag]++;
      }
      const variance = this._computeAntiAffinityVariance(taggedHosts);
      for (const tag of vmTags) {
        sourceHost.tags[tag]++;
        destinationHost.tags[tag]--;
      }
      if (variance < bestVariance) {
        if (vm.xenTools) {
          bestVariance = variance;
          bestVm = vm;
        } else {
          debugAffinity(`VM (${vm.id}) of Host (${sourceHost.id}) does not support pool migration.`);
        }
      }
    }
    return bestVm;
  }
}
exports.default = Plan;
//# sourceMappingURL=plan.js.map