"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.PrivateNetwork = void 0;
var _assert = _interopRequireDefault(require("assert"));
var _log = require("@xen-orchestra/log");
var _lodash = require("lodash");
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const log = (0, _log.createLogger)('xo:xo-server:sdn-controller:private-network');
const CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789?!';
const createPassword = () => Array.from({
  length: 16
}, _ => (0, _lodash.sample)(CHARS)).join('');
class PrivateNetwork {
  constructor(controller, uuid, preferredCenter) {
    this._preferredCenter = preferredCenter;
    this.controller = controller;
    this.uuid = uuid;
    this.networks = {};
    this.controller.registerPrivateNetwork(this);
  }
  async addHost(host) {
    var _this$center, _otherConfig$xoSdnC, _otherConfig$xoSdnC2;
    if (host.$ref === ((_this$center = this.center) === null || _this$center === void 0 ? void 0 : _this$center.$ref)) {
      return;
    }
    const hostClient = this.controller.ovsdbClients[host.$ref];
    if (hostClient === undefined) {
      log.error('No OVSDB client found', {
        host: host.name_label,
        pool: host.$pool.name_label
      });
      return;
    }
    const centerClient = this.controller.ovsdbClients[this.center.$ref];
    if (centerClient === undefined) {
      log.error('No OVSDB client found for star-center', {
        privateNetwork: this.uuid,
        host: this.center.name_label,
        pool: this.center.$pool.name_label
      });
      return;
    }
    const network = this.networks[host.$pool.uuid];
    const centerNetwork = this.networks[this.center.$pool.uuid];
    const otherConfig = network.other_config;
    const encapsulation = (_otherConfig$xoSdnC = otherConfig['xo:sdn-controller:encapsulation']) !== null && _otherConfig$xoSdnC !== void 0 ? _otherConfig$xoSdnC : 'gre';
    const vni = (_otherConfig$xoSdnC2 = otherConfig['xo:sdn-controller:vni']) !== null && _otherConfig$xoSdnC2 !== void 0 ? _otherConfig$xoSdnC2 : '0';
    const password = otherConfig['xo:sdn-controller:encrypted'] === 'true' ? createPassword() : undefined;
    const pifDevice = otherConfig['xo:sdn-controller:pif-device'];
    const pifVlan = +otherConfig['xo:sdn-controller:vlan'];
    const hostPif = hostClient.host.$PIFs.find(pif => (pif === null || pif === void 0 ? void 0 : pif.device) === pifDevice && pif.VLAN === pifVlan && pif.ip_configuration_mode !== 'None');
    const centerDevice = centerNetwork.other_config['xo:sdn-controller:pif-device'];
    const centerVlan = +centerNetwork.other_config['xo:sdn-controller:vlan'];
    const centerPif = centerClient.host.$PIFs.find(pif => (pif === null || pif === void 0 ? void 0 : pif.device) === centerDevice && pif.VLAN === centerVlan && pif.ip_configuration_mode !== 'None');
    (0, _assert.default)(hostPif !== undefined, 'No PIF found', {
      privateNetwork: this.uuid,
      pifDevice,
      pifVlan,
      host: host.name_label
    });
    (0, _assert.default)(centerPif !== undefined, 'No PIF found in center', {
      privateNetwork: this.uuid,
      pifDevice,
      pifVlan,
      host: this.center.name_label
    });
    let bridgeName;
    try {
      ;
      [bridgeName] = await Promise.all([hostClient.addInterfaceAndPort(network, centerPif.IP, encapsulation, vni, password, this.uuid), centerClient.addInterfaceAndPort(centerNetwork, hostPif.IP, encapsulation, vni, password, this.uuid)]);
    } catch (error) {
      log.error('Error while connecting host to private network', {
        error,
        privateNetwork: this.uuid,
        network: network.name_label,
        host: host.name_label,
        pool: host.$pool.name_label
      });
      return;
    }
    log.info('Host added', {
      privateNetwork: this.uuid,
      network: network.name_label,
      host: host.name_label,
      pool: host.$pool.name_label
    });
    return bridgeName;
  }
  addNetwork(network) {
    this.networks[network.$pool.uuid] = network;
    log.info('Adding network', {
      privateNetwork: this.uuid,
      network: network.name_label,
      pool: network.$pool.name_label
    });
    if (this.center === undefined) {
      return this.electNewCenter();
    }
    const hosts = (0, _lodash.filter)(network.$pool.$xapi.objects.all, {
      $type: 'host'
    });
    return Promise.all(hosts.map(async host => {
      const hostClient = this.controller.ovsdbClients[host.$ref];
      const network = this.networks[host.$pool.uuid];
      await hostClient.resetForNetwork(network, this.uuid);
      await this.addHost(host);
    }));
  }
  async electNewCenter() {
    delete this.center;
    if (this._preferredCenter !== undefined) {
      this._preferredCenter = await this._preferredCenter.$xapi.barrier(this._preferredCenter.$ref);
    }
    this.center = this._findBestCenter();
    if (this.center === undefined) {
      log.error('No available host to elect new star-center', {
        privateNetwork: this.uuid
      });
      return;
    }
    await this._reset();
    await Promise.all(this._getHosts().map(host => this.addHost(host)));
    log.info('New star-center elected', {
      center: this.center.name_label,
      privateNetwork: this.uuid
    });
  }
  getPools() {
    const pools = [];
    (0, _lodash.forOwn)(this.networks, network => {
      pools.push(network.$pool);
    });
    return pools;
  }
  _reset() {
    return Promise.all(this._getHosts().map(async host => {
      const hostClient = this.controller.ovsdbClients[host.$ref];
      if (hostClient === undefined) {
        return;
      }
      const network = this.networks[host.$pool.uuid];
      try {
        await hostClient.resetForNetwork(network, this.uuid);
      } catch (error) {
        log.error('Error while resetting private network', {
          error,
          privateNetwork: this.uuid,
          network: network.name_label,
          host: host.name_label,
          pool: network.$pool.name_label
        });
      }
    }));
  }
  _getHosts() {
    const hosts = [];
    (0, _lodash.forOwn)(this.networks, network => {
      hosts.push(...(0, _lodash.filter)(network.$pool.$xapi.objects.all, {
        $type: 'host'
      }));
    });
    return hosts;
  }
  _hostCanBeCenter(host) {
    const pif = host.$PIFs.find(_ => _.network === this.networks[host.$pool.uuid].$ref);
    return (pif === null || pif === void 0 ? void 0 : pif.currently_attached) && host.$metrics.live;
  }
  _findBestCenter() {
    if (this._preferredCenter !== undefined && this._hostCanBeCenter(this._preferredCenter)) {
      return this._preferredCenter;
    }
    for (const host of this._getHosts()) {
      if (this._hostCanBeCenter(host)) {
        return host;
      }
    }
  }
}
exports.PrivateNetwork = PrivateNetwork;
//# sourceMappingURL=private-network.js.map