"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.OpenFlowChannel = void 0;
var _ipaddr = _interopRequireDefault(require("ipaddr.js"));
var _openflow = _interopRequireDefault(require("@xen-orchestra/openflow"));
var _parseSocket = _interopRequireDefault(require("@xen-orchestra/openflow/parse-socket"));
var _coalesceCalls = require("@vates/coalesce-calls");
var _log = require("@xen-orchestra/log");
var _events = require("events");
var _promiseToolbox = require("promise-toolbox");
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const log = (0, _log.createLogger)('xo:xo-server:sdn-controller:openflow-controller');
const version = _openflow.default.versions.openFlow11;
const ofProtocol = _openflow.default.protocols[version];
const OPENFLOW_PORT = ofProtocol.sslPort;
const parseIp = ipAddress => {
  if (ipAddress === '') {
    return;
  }
  let addr, mask;
  if (ipAddress.includes('/')) {
    const ip = _ipaddr.default.parseCIDR(ipAddress);
    addr = ip[0].toString();
    const maskOctets = _ipaddr.default.IPv4.subnetMaskFromPrefixLength(ip[1]).octets;
    mask = _ipaddr.default.fromByteArray(maskOctets.map(i => 255 - i)).toString();
  } else {
    const ip = _ipaddr.default.parse(ipAddress);
    addr = ip.toString();
  }
  return {
    addr,
    mask
  };
};
const dlAndNwProtocolFromString = protocol => {
  switch (protocol) {
    case 'IP':
      return {
        dlType: ofProtocol.dlType.ip
      };
    case 'ICMP':
      return {
        dlType: ofProtocol.dlType.ip,
        nwProto: ofProtocol.nwProto.icmp
      };
    case 'TCP':
      return {
        dlType: ofProtocol.dlType.ip,
        nwProto: ofProtocol.nwProto.tcp
      };
    case 'UDP':
      return {
        dlType: ofProtocol.dlType.ip,
        nwProto: ofProtocol.nwProto.udp
      };
    case 'ARP':
      return {
        dlType: ofProtocol.dlType.arp
      };
    default:
      return {};
  }
};
class OpenFlowChannel extends _events.EventEmitter {
  constructor(host, tlsHelper) {
    super();
    this.host = host;
    this._tlsHelper = tlsHelper;
    this._coalesceConnect = (0, _coalesceCalls.coalesceCalls)(this._connect);
    this._socket = undefined;
    log.debug('New OpenFlow channel', {
      host: this.host.name_label
    });
  }
  async addRule(vif, allow, protocol, port, ipRange, direction, ofport) {
    log.info('Adding OF rule', {
      allow,
      protocol,
      port,
      ipRange,
      direction,
      vif: vif.uuid
    });
    const instructions = [{
      type: ofProtocol.instructionType.applyActions,
      actions: allow ? [{
        type: ofProtocol.actionType.output,
        port: ofProtocol.port.normal
      }] : []
    }];
    const ip = parseIp(ipRange);
    const {
      dlType,
      nwProto
    } = dlAndNwProtocolFromString(protocol);
    const mac = vif.MAC;
    await this._coalesceConnect();
    if (direction.includes('from')) {
      this._addFlow({
        type: ofProtocol.matchType.standard,
        dl_type: dlType,
        dl_src: mac,
        nw_proto: nwProto,
        nw_dst: ip === null || ip === void 0 ? void 0 : ip.addr,
        nw_dst_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
        tp_src: port,
        in_port: ofport
      }, instructions);
      if (nwProto !== undefined) {
        this._addFlow({
          type: ofProtocol.matchType.standard,
          dl_type: dlType,
          dl_dst: mac,
          nw_proto: nwProto,
          nw_src: ip === null || ip === void 0 ? void 0 : ip.addr,
          nw_src_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
          tp_dst: port
        }, instructions);
      }
    }
    if (direction.includes('to')) {
      if (nwProto !== undefined) {
        this._addFlow({
          type: ofProtocol.matchType.standard,
          dl_type: dlType,
          dl_src: mac,
          nw_proto: nwProto,
          nw_dst: ip === null || ip === void 0 ? void 0 : ip.addr,
          nw_dst_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
          tp_dst: port,
          in_port: ofport
        }, instructions);
      }
      this._addFlow({
        type: ofProtocol.matchType.standard,
        dl_type: dlType,
        dl_dst: mac,
        nw_proto: nwProto,
        nw_src: ip === null || ip === void 0 ? void 0 : ip.addr,
        nw_src_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
        tp_src: port
      }, instructions);
    }
  }
  async deleteRule(vif, protocol, port, ipRange, direction, ofport) {
    log.info('Deleting OF rule', {
      protocol,
      port,
      ipRange,
      direction,
      vif: vif.uuid
    });
    const ip = parseIp(ipRange);
    const {
      dlType,
      nwProto
    } = dlAndNwProtocolFromString(protocol);
    const mac = vif.MAC;
    await this._coalesceConnect();
    if (direction.includes('from')) {
      this._removeFlows({
        type: ofProtocol.matchType.standard,
        dl_type: dlType,
        dl_src: mac,
        nw_proto: nwProto,
        nw_dst: ip === null || ip === void 0 ? void 0 : ip.addr,
        nw_dst_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
        tp_src: port
      });
      if (nwProto !== undefined) {
        this._removeFlows({
          type: ofProtocol.matchType.standard,
          dl_type: dlType,
          dl_dst: mac,
          nw_proto: nwProto,
          nw_src: ip === null || ip === void 0 ? void 0 : ip.addr,
          nw_src_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
          tp_dst: port
        });
      }
    }
    if (direction.includes('to')) {
      if (nwProto !== undefined) {
        this._removeFlows({
          type: ofProtocol.matchType.standard,
          dl_type: dlType,
          dl_src: mac,
          nw_proto: nwProto,
          nw_dst: ip === null || ip === void 0 ? void 0 : ip.addr,
          nw_dst_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
          tp_dst: port
        });
      }
      this._removeFlows({
        type: ofProtocol.matchType.standard,
        dl_type: dlType,
        dl_dst: mac,
        nw_proto: nwProto,
        nw_src: ip === null || ip === void 0 ? void 0 : ip.addr,
        nw_src_mask: ip === null || ip === void 0 ? void 0 : ip.mask,
        tp_src: port
      });
    }
  }
  _processMessage(message) {
    if (message.header === undefined) {
      log.error('Failed to get header while processing message', {
        message
      });
      return;
    }
    const ofType = message.header.type;
    switch (ofType) {
      case ofProtocol.type.hello:
        this._sendPacket(this._syncMessage(ofProtocol.type.hello, message.header.xid));
        this._sendPacket(this._syncMessage(ofProtocol.type.featuresRequest, message.header.xid));
        break;
      case ofProtocol.type.error:
        {
          const {
            code,
            type
          } = message;
          log.error('OpenFlow error', {
            code,
            type
          });
        }
        break;
      case ofProtocol.type.echoRequest:
        this._sendPacket(this._syncMessage(ofProtocol.type.echoReply, message.header.xid));
        break;
      case ofProtocol.type.packetIn:
        log.debug('PACKET_IN');
        break;
      case ofProtocol.type.featuresReply:
        {
          const {
            datapath_id: dpid,
            capabilities,
            ports
          } = message;
          log.debug('FEATURES_REPLY', {
            dpid,
            capabilities,
            ports
          });
          this._sendPacket(this._syncMessage(ofProtocol.type.getConfigRequest, message.header.xid));
        }
        break;
      case ofProtocol.type.getConfigReply:
        {
          const {
            flags
          } = message;
          log.debug('CONFIG_REPLY', {
            flags
          });
          this.emit('ofConnected');
        }
        break;
      case ofProtocol.type.portStatus:
        log.debug('PORT_STATUS');
        break;
      case ofProtocol.type.flowRemoved:
        log.debug('FLOW_REMOVED');
        break;
      default:
        log.error('Unknown OpenFlow type', {
          ofType
        });
        break;
    }
  }
  _addFlow(match, instructions) {
    const packet = this._flowModMessage(ofProtocol.flowModCommand.add, match, instructions);
    this._sendPacket(packet);
  }
  _removeFlows(match) {
    const packet = this._flowModMessage(ofProtocol.flowModCommand.delete, match);
    this._sendPacket(packet);
  }
  _syncMessage(type, xid = 1) {
    return {
      header: {
        version,
        type,
        xid
      }
    };
  }
  _flowModMessage(command, match, instructions = []) {
    return {
      ...this._syncMessage(ofProtocol.type.flowMod),
      command,
      flags: ofProtocol.flowModFlags.sendFlowRem,
      match,
      instructions
    };
  }
  _sendPacket(packet) {
    const buf = _openflow.default.pack(packet);
    try {
      this._socket.write(buf);
    } catch (error) {
      log.error('Error while writing into socket', {
        error,
        host: this.host.name_label
      });
    }
  }
  async _parseMessages() {
    for await (const msg of (0, _parseSocket.default)(this._socket)) {
      if (msg.header !== undefined) {
        this._processMessage(msg);
      } else {
        log.error('Error: Message is unparseable', {
          msg
        });
      }
    }
  }
  async _connect() {
    if (this._socket !== undefined) {
      return;
    }
    this._socket = await this._tlsHelper.connect(this.host.address, OPENFLOW_PORT);
    const deleteSocket = () => {
      this._socket = undefined;
    };
    this._socket.on('error', deleteSocket);
    this._socket.on('end', deleteSocket);
    this._parseMessages().catch(error => {
      log.error('Error while parsing OF messages', error);
    });
    await (0, _promiseToolbox.fromEvent)(this, 'ofConnected');
  }
}
exports.OpenFlowChannel = OpenFlowChannel;
//# sourceMappingURL=openflow-channel.js.map