import * as multiparty from 'multiparty';
import assert from 'assert';
import getStream from 'get-stream';
import { createLogger } from '@xen-orchestra/log';
import { defer } from 'golike-defer';
import { format, JsonRpcError } from 'json-rpc-peer';
import { noSuchObject } from 'xo-common/api-errors.js';
import { pipeline } from 'stream';
import { peekFooterFromVhdStream } from 'vhd-lib';
import { vmdkToVhd } from 'xo-vmdk-to-vhd';
import { VDI_FORMAT_VHD, VDI_FORMAT_RAW } from '../xapi/index.mjs';
import { parseSize } from '../utils.mjs';
import { readChunk } from '@vates/read-chunk';
const log = createLogger('xo:disk');
export const create = defer(async function ($defer, {
  name,
  size,
  sr,
  vm,
  bootable,
  position,
  mode
}) {
  const attach = vm !== undefined;
  do {
    let resourceSet;
    if (attach && (resourceSet = vm.resourceSet) != null) {
      try {
        await this.checkResourceSetConstraints(resourceSet, this.apiContext.user.id, [sr.id]);
        await this.allocateLimitsInResourceSet({
          disk: size
        }, resourceSet);
        $defer.onFailure(() => this.releaseLimitsInResourceSet({
          disk: size
        }, resourceSet));
        break;
      } catch (error) {
        if (!noSuchObject.is(error, {
          id: resourceSet
        })) {
          throw error;
        }
      }
    }
    await this.checkPermissions([[sr.id, 'administrate']]);
  } while (false);
  const xapi = this.getXapi(sr);
  const vdi = await xapi._getOrWaitObject(await xapi.VDI_create({
    name_label: name,
    SR: sr._xapiRef,
    virtual_size: parseSize(size)
  }));
  $defer.onFailure(() => vdi.$destroy());
  if (attach) {
    await xapi.VBD_create({
      bootable,
      mode,
      userdevice: position,
      VDI: vdi.$ref,
      VM: vm._xapiRef
    });
  }
  return vdi.$id;
});
create.description = 'create a new disk on a SR';
create.params = {
  name: {
    type: 'string'
  },
  size: {
    type: ['integer', 'string']
  },
  sr: {
    type: 'string'
  },
  vm: {
    type: 'string',
    optional: true
  },
  bootable: {
    type: 'boolean',
    optional: true
  },
  mode: {
    type: 'string',
    optional: true
  },
  position: {
    type: 'string',
    optional: true
  }
};
create.resolve = {
  vm: ['vm', 'VM', 'administrate'],
  sr: ['sr', 'SR', false]
};
const VHD = 'vhd';
const VMDK = 'vmdk';
async function handleExportContent(req, res, {
  filename,
  format,
  vdi
}) {
  const stream = format === VMDK ? await vdi.$xapi.exportVdiAsVmdk(vdi.$id, filename) : await vdi.$exportContent({
    format: VDI_FORMAT_VHD
  });
  req.on('close', () => stream.destroy());
  const {
    headers = {},
    length,
    statusCode = 200,
    statusMessage = 'OK'
  } = stream;
  headers['content-disposition'] = 'attachment';
  if (headers['content-length'] === undefined && length !== undefined) {
    headers['content-length'] = length;
  }
  res.writeHead(statusCode, statusMessage != null ? statusMessage : '', headers);
  pipeline(stream, res, error => {
    if (error != null) {
      log.warn('disk.exportContent', {
        error
      });
    }
  });
}
export async function exportContent({
  vdi,
  format = VHD
}) {
  const filename = (vdi.name_label || 'unknown') + '.' + (format === VHD ? 'vhd' : 'vmdk');
  return {
    $getFrom: await this.registerHttpRequest(handleExportContent, {
      vdi: this.getXapiObject(vdi),
      filename,
      format
    }, {
      suffix: `/${encodeURIComponent(filename)}`
    })
  };
}
exportContent.description = 'export the content of a VDI';
exportContent.params = {
  id: {
    type: 'string'
  },
  format: {
    enum: [VMDK, VHD],
    optional: true
  }
};
exportContent.resolve = {
  vdi: ['id', ['VDI', 'VDI-snapshot'], 'view']
};
async function handleImportContent(req, res, {
  vdi
}) {
  req.setTimeout(43200000);
  req.length = +req.headers['content-length'];
  await vdi.$importContent(req, {
    format: VDI_FORMAT_VHD
  });
  res.end(format.response(0, true));
}
export async function importContent({
  vdi
}) {
  return {
    $sendTo: await this.registerHttpRequest(handleImportContent, {
      vdi: this.getXapiObject(vdi)
    })
  };
}
importContent.description = 'import contents into a VDI';
importContent.params = {
  id: {
    type: 'string'
  }
};
importContent.resolve = {
  vdi: ['id', ['VDI'], 'operate']
};
async function handleImport(req, res, {
  type,
  name,
  description,
  vmdkData,
  srId,
  xapi
}) {
  req.setTimeout(43200000);
  req.length = req.headers['content-length'];
  let vhdStream, size;
  await new Promise((resolve, reject) => {
    const promises = [];
    const form = new multiparty.Form();
    form.on('error', reject);
    form.on('part', async part => {
      try {
        if (part.name !== 'file') {
          promises.push((async () => {
            const buffer = await getStream.buffer(part);
            vmdkData[part.name] = new Uint32Array(buffer.buffer, buffer.byteOffset, buffer.length / Uint32Array.BYTES_PER_ELEMENT);
          })());
        } else {
          let diskFormat = VDI_FORMAT_VHD;
          await Promise.all(promises);
          part.length = part.byteCount;
          switch (type) {
            case 'vmdk':
              vhdStream = await vmdkToVhd(part, vmdkData.grainLogicalAddressList, vmdkData.grainFileOffsetList);
              size = vmdkData.capacity;
              break;
            case 'vhd':
              {
                const footer = await peekFooterFromVhdStream(part).catch(e => {
                  if (e instanceof assert.AssertionError) {
                    throw new JsonRpcError(`Vhd file had an invalid header ${e}`);
                  }
                  throw e;
                });
                vhdStream = part;
                size = footer.currentSize;
              }
              break;
            case 'iso':
              diskFormat = VDI_FORMAT_RAW;
              vhdStream = part;
              size = part.byteCount;
              break;
            default:
              throw new JsonRpcError(`Unknown disk type, expected "iso", "vhd" or "vmdk", got ${type}`);
          }
          const vdi = await xapi._getOrWaitObject(await xapi.VDI_create({
            name_description: description,
            name_label: name,
            SR: xapi.getObject(srId, 'SR').$ref,
            virtual_size: parseSize(size)
          }));
          try {
            await vdi.$importContent(vhdStream, {
              format: diskFormat
            });
            let buffer;
            const CHUNK_SIZE = 1024 * 1024;
            do {
              var _buffer;
              buffer = await readChunk(part, CHUNK_SIZE);
            } while (((_buffer = buffer) === null || _buffer === void 0 ? void 0 : _buffer.length) === CHUNK_SIZE);
            res.end(format.response(0, vdi.$id));
          } catch (e) {
            await vdi.$destroy();
            throw e;
          }
          resolve();
        }
      } catch (e) {
        res.writeHead(500);
        res.end(format.error(0, new JsonRpcError(e.message)));
        req.destroy();
        reject(e);
      }
    });
    form.parse(req);
  });
}
async function importDisk({
  sr,
  type,
  name,
  description,
  vmdkData
}) {
  return {
    $sendTo: await this.registerHttpRequest(handleImport, {
      description,
      name,
      srId: sr._xapiId,
      type,
      vmdkData,
      xapi: this.getXapi(sr)
    })
  };
}
export { importDisk as import };
importDisk.params = {
  description: {
    type: 'string',
    minLength: 0,
    optional: true
  },
  name: {
    type: 'string'
  },
  sr: {
    type: 'string'
  },
  type: {
    type: 'string'
  },
  vmdkData: {
    type: 'object',
    optional: true,
    properties: {
      capacity: {
        type: 'integer'
      }
    }
  }
};
importDisk.resolve = {
  sr: ['sr', 'SR', 'administrate']
};
//# sourceMappingURL=disk.mjs.map