import Ajv from 'ajv';
import cloneDeep from 'lodash/cloneDeep.js';
import mapToArray from 'lodash/map.js';
import noop from 'lodash/noop.js';
import { createLogger } from '@xen-orchestra/log';
import { invalidParameters, noSuchObject } from 'xo-common/api-errors.js';
import * as sensitiveValues from '../sensitive-values.mjs';
import { PluginsMetadata } from '../models/plugin-metadata.mjs';
const log = createLogger('xo:xo-mixins:plugins');
export default class {
  constructor(app) {
    this._ajv = new Ajv({
      strict: 'log',
      useDefaults: true
    }).addVocabulary(['$multiline', '$type', 'enumNames']);
    this._plugins = {
      __proto__: null
    };
    this._pluginsMetadata = new PluginsMetadata({
      connection: app._redis,
      namespace: 'plugin-metadata'
    });
    app.hooks.on('start', () => {
      app.addConfigManager('plugins', () => this._pluginsMetadata.get(), plugins => Promise.all(plugins.map(async plugin => {
        await this._pluginsMetadata.save(plugin);
        if (plugin.configuration !== undefined && this._plugins[plugin.id] !== undefined) {
          await this.configurePlugin(plugin.id, plugin.configuration);
        }
      })));
    });
  }
  _getRawPlugin(id) {
    const plugin = this._plugins[id];
    if (!plugin) {
      throw noSuchObject(id, 'plugin');
    }
    return plugin;
  }
  _getPluginMetadata(id) {
    return this._pluginsMetadata.first(id);
  }
  async registerPlugin(name, instance, configurationSchema, configurationPresets, description, keywords, testSchema, version) {
    const id = name;
    const plugin = this._plugins[id] = {
      configurationPresets,
      configurationSchema,
      configured: configurationSchema === undefined,
      description,
      id,
      instance,
      keywords,
      name,
      testable: typeof instance.test === 'function',
      testSchema,
      unloadable: typeof instance.unload === 'function',
      version
    };
    const metadata = await this._getPluginMetadata(id);
    let autoload = true;
    let configuration;
    if (metadata !== undefined) {
      ;
      ({
        autoload,
        configuration
      } = metadata);
    } else {
      log.info(`[NOTICE] register plugin ${name} for the first time`);
      await this._pluginsMetadata.save({
        id,
        autoload
      });
    }
    if (!plugin.configured) {
      const tryEmptyConfig = configuration === undefined;
      try {
        await this._configurePlugin(plugin, tryEmptyConfig ? {} : configuration);
      } catch (error) {
        if (tryEmptyConfig) {
          return;
        }
        throw error;
      }
    }
    if (autoload) {
      await this.loadPlugin(id);
    }
  }
  async getPlugin(id) {
    const {
      configurationPresets,
      configurationSchema,
      description,
      keywords,
      loaded,
      name,
      testable,
      testSchema,
      unloadable,
      version
    } = this._getRawPlugin(id);
    const {
      autoload,
      configuration
    } = (await this._getPluginMetadata(id)) || {};
    return {
      id,
      name,
      autoload,
      description,
      loaded,
      keywords,
      unloadable,
      version,
      configuration: sensitiveValues.obfuscate(configuration),
      configurationPresets,
      configurationSchema,
      testable,
      testSchema
    };
  }
  async getOptionalPlugin(id) {
    try {
      return await this.getPlugin(id);
    } catch (error) {
      if (!noSuchObject.is(error, {
        id,
        type: 'plugin'
      })) {
        throw error;
      }
    }
  }
  async getPlugins() {
    return Promise.all(mapToArray(this._plugins, ({
      id
    }) => this.getPlugin(id)));
  }
  async _configurePlugin(plugin, configuration) {
    const {
      configurationSchema
    } = plugin;
    if (!configurationSchema) {
      throw invalidParameters('plugin not configurable');
    }
    const validate = this._ajv.compile(configurationSchema);
    configuration = cloneDeep(configuration);
    if (!validate(configuration)) {
      throw invalidParameters(validate.errors);
    }
    await plugin.instance.configure(configuration, {
      loaded: plugin.loaded
    });
    plugin.configured = true;
  }
  async configurePlugin(id, configuration) {
    const plugin = this._getRawPlugin(id);
    const metadata = await this._getPluginMetadata(id);
    if (metadata !== undefined) {
      configuration = sensitiveValues.merge(configuration, metadata.configuration);
    }
    await this._configurePlugin(plugin, configuration);
    await this._pluginsMetadata.merge(id, {
      configuration
    });
  }
  async disablePluginAutoload(id) {
    await this._pluginsMetadata.merge(id, {
      autoload: false
    });
  }
  async enablePluginAutoload(id) {
    await this._pluginsMetadata.merge(id, {
      autoload: true
    });
  }
  async loadPlugin(id) {
    const plugin = this._getRawPlugin(id);
    if (plugin.loaded) {
      throw invalidParameters('plugin already loaded');
    }
    if (!plugin.configured) {
      throw invalidParameters('plugin not configured');
    }
    if (plugin.loading) {
      throw invalidParameters('plugin is loading');
    }
    plugin.loading = true;
    try {
      await plugin.instance.load();
      plugin.loaded = true;
    } finally {
      plugin.loading = false;
    }
  }
  async unloadPlugin(id) {
    const plugin = this._getRawPlugin(id);
    if (!plugin.loaded) {
      throw invalidParameters('plugin already unloaded');
    }
    if (plugin.unloadable === false) {
      throw invalidParameters('plugin cannot be unloaded');
    }
    await plugin.instance.unload();
    plugin.loaded = false;
  }
  async purgePluginConfiguration(id) {
    await this.unloadPlugin(id).catch(noop);
    await this._pluginsMetadata.merge(id, {
      configuration: undefined
    });
  }
  async testPlugin(id, data) {
    const plugin = this._getRawPlugin(id);
    if (!plugin.testable) {
      throw invalidParameters('plugin not testable');
    }
    if (!plugin.loaded) {
      throw invalidParameters('plugin not loaded');
    }
    const {
      testSchema
    } = plugin;
    if (testSchema) {
      if (data == null) {
        throw invalidParameters([{
          field: 'data',
          message: 'is the wrong type'
        }]);
      }
      const validate = this._ajv.compile(testSchema);
      if (!validate(data)) {
        throw invalidParameters(validate.errors);
      }
    }
    await plugin.instance.test(data);
  }
}
//# sourceMappingURL=plugins.mjs.map