import assert from 'node:assert/strict';
import filter from 'lodash/filter.js';
import { createLogger } from '@xen-orchestra/log';
import { ignoreErrors } from 'promise-toolbox';
import { hash, needsRehash, verify } from 'hashy';
import { invalidCredentials, noSuchObject } from 'xo-common/api-errors.js';
import * as XenStore from '../_XenStore.mjs';
import { Groups } from '../models/group.mjs';
import { Users } from '../models/user.mjs';
import { forEach, isEmpty, lightSet } from '../utils.mjs';
const log = createLogger('xo:xo-mixins:subjects');
const addToArraySet = (set, value) => set !== undefined ? set.includes(value) ? set : set.concat(value) : [value];
const removeFromArraySet = (set, value) => set && filter(set, current => current !== value);
export default class {
  constructor(app) {
    this._app = app;
    const redis = app._redis;
    const groupsDb = this._groups = new Groups({
      connection: redis,
      namespace: 'group'
    });
    const usersDb = this._users = new Users({
      connection: redis,
      namespace: 'user',
      indexes: ['email']
    });
    app.hooks.on('clean', () => Promise.all([groupsDb.rebuildIndexes(), usersDb.rebuildIndexes()]));
    app.hooks.on('start', async () => {
      app.addConfigManager('groups', () => groupsDb.get(), groups => Promise.all(groups.map(group => groupsDb.save(group))), ['users']);
      app.addConfigManager('users', () => usersDb.get(), users => Promise.all(users.map(async user => {
        const userId = user.id;
        const conflictUsers = await usersDb.get({
          email: user.email
        });
        if (!isEmpty(conflictUsers)) {
          await Promise.all(conflictUsers.map(({
            id
          }) => id !== userId && this.deleteUser(id)));
        }
        return usersDb.save(user);
      })));
      if (!(await usersDb.exists())) {
        const key = 'vm-data/admin-account';
        const {
          email = 'admin@admin.net',
          password = 'admin'
        } = await XenStore.read(key).then(JSON.parse).catch(() => ({}));
        await this.createUser({
          email,
          password,
          permission: 'admin'
        });
        log.info(`Default user created: ${email} with password ${password}`);
        ignoreErrors.call(XenStore.rm(key));
      }
    });
  }
  async createUser({
    name,
    password,
    ...properties
  }) {
    if (name) {
      properties.email = name;
    }
    if (password) {
      properties.pw_hash = await hash(password);
    }
    const user = await this._users.create(properties);
    return user;
  }
  async deleteUser(id) {
    var _context;
    const user = await this.getUser(id);
    await this._users.remove(id);
    (_context = this._app.getAuthenticationTokensForUser(id).then(tokens => {
      forEach(tokens, token => {
        var _context2;
        (_context2 = this._app.deleteAuthenticationToken(id), ignoreErrors).call(_context2);
      });
    }), ignoreErrors).call(_context);
    this._app.getAclsForSubject(id).then(acls => {
      forEach(acls, acl => {
        var _context3;
        (_context3 = this._app.removeAcl(id, acl.object, acl.action), ignoreErrors).call(_context3);
      });
    });
    forEach(user.groups, groupId => {
      var _context4;
      (_context4 = this.getGroup(groupId).then(group => this._removeUserFromGroup(id, group)), ignoreErrors).call(_context4);
    });
  }
  async updateUser(id, {
    email,
    authProviders,
    name = email,
    password,
    permission,
    preferences
  }) {
    const user = await this.getUser(id);
    if (name) {
      user.name = name;
    }
    if (permission) {
      user.permission = permission;
    }
    if (password) {
      user.pw_hash = await hash(password);
    }
    const newPreferences = {
      ...user.preferences
    };
    forEach(preferences, (value, name) => {
      if (value == null) {
        delete newPreferences[name];
      } else {
        newPreferences[name] = value;
      }
    });
    user.preferences = isEmpty(newPreferences) ? undefined : newPreferences;
    const newAuthProviders = {
      ...user.authProviders
    };
    forEach(authProviders, (value, name) => {
      if (value == null) {
        delete newAuthProviders[name];
      } else {
        newAuthProviders[name] = value;
      }
    });
    user.authProviders = isEmpty(newAuthProviders) ? undefined : newAuthProviders;
    user.email = user.name;
    delete user.name;
    await this._users.save(user);
  }
  async _getUser(id) {
    const user = await this._users.first(id);
    if (user === undefined) {
      throw noSuchObject(id, 'user');
    }
    return user;
  }
  async getUser(id) {
    const user = await this._getUser(id);
    user.name = user.email;
    return user;
  }
  async getAllUsers() {
    return this._users.get();
  }
  async getUserByName(username, returnNullIfMissing) {
    const user = await this._users.first({
      email: username
    });
    if (user !== undefined) {
      return user;
    }
    if (returnNullIfMissing) {
      return null;
    }
    throw noSuchObject(username, 'user');
  }
  async registerUser(provider, name) {
    const user = await this.getUserByName(name, true);
    if (user) {
      if (user._provider !== provider) {
        throw new Error(`the name ${name} is already taken`);
      }
      return user;
    }
    if (!this._app.config.get('createUserOnFirstSignin')) {
      throw new Error(`registering ${name} user is forbidden`);
    }
    return this.createUser({
      name,
      _provider: provider
    });
  }
  async registerUser2(providerId, {
    user: {
      id,
      name
    },
    data
  }) {
    var _user;
    assert.equal(typeof name, 'string');
    assert.notEqual(name, '');
    const users = await this.getAllUsers();
    let user = users.find(user => {
      var _user$authProviders, _user$authProviders$p;
      return ((_user$authProviders = user.authProviders) === null || _user$authProviders === void 0 ? void 0 : (_user$authProviders$p = _user$authProviders[providerId]) === null || _user$authProviders$p === void 0 ? void 0 : _user$authProviders$p.id) === id;
    });
    let conflictingUser;
    if (((_user = user) === null || _user === void 0 ? void 0 : _user.email) !== name) {
      conflictingUser = users.find(user => user.email === name);
      if (conflictingUser !== undefined) {
        if (!this._app.config.get('authentication.mergeProvidersUsers')) {
          throw new Error(`User with username ${name} already exists`);
        }
        if (user !== undefined) {} else {
          user = conflictingUser;
        }
      }
    } else if (user !== undefined) {
      return user;
    }
    if (user === undefined) {
      if (!this._app.config.get('createUserOnFirstSignin')) {
        throw new Error(`registering ${name} user is forbidden`);
      }
      user = await this.createUser({
        name,
        authProviders: {
          [providerId]: {
            id,
            data
          }
        }
      });
    } else {
      var _user$authProviders2, _user$authProviders2$;
      await this.updateUser(user.id, {
        name: (user.authProviders === undefined || Object.keys(user.authProviders).length < 2) && conflictingUser === undefined ? name : undefined,
        authProviders: {
          ...user.authProviders,
          [providerId]: {
            id,
            data: data !== undefined ? data : (_user$authProviders2 = user.authProviders) === null || _user$authProviders2 === void 0 ? void 0 : (_user$authProviders2$ = _user$authProviders2[providerId]) === null || _user$authProviders2$ === void 0 ? void 0 : _user$authProviders2$.data
          }
        }
      });
    }
    return this.getUser(user.id);
  }
  async changeUserPassword(userId, oldPassword, newPassword) {
    if (!(await this.checkUserPassword(userId, oldPassword, false))) {
      throw invalidCredentials();
    }
    await this.updateUser(userId, {
      password: newPassword
    });
  }
  async checkUserPassword(userId, password, updateIfNecessary = true) {
    const {
      pw_hash: hash
    } = await this.getUser(userId);
    if (!(hash && (await verify(password, hash)))) {
      return false;
    }
    if (updateIfNecessary && needsRehash(hash)) {
      await this.updateUser(userId, {
        password
      });
    }
    return true;
  }
  async createGroup({
    name,
    provider,
    providerGroupId
  }) {
    const group = await this._groups.create(name, provider, providerGroupId);
    return group;
  }
  async deleteGroup(id) {
    const group = await this.getGroup(id);
    await this._groups.remove(id);
    this._app.getAclsForSubject(id).then(acls => {
      forEach(acls, acl => {
        var _context5;
        (_context5 = this._app.removeAcl(id, acl.object, acl.action), ignoreErrors).call(_context5);
      });
    });
    forEach(group.users, userId => {
      var _context6;
      (_context6 = this.getUser(userId).then(user => this._removeGroupFromUser(id, user)), ignoreErrors).call(_context6);
    });
  }
  async updateGroup(id, {
    name
  }) {
    const group = await this.getGroup(id);
    if (name) group.name = name;
    await this._groups.save(group);
  }
  async getGroup(id) {
    const group = await this._groups.first(id);
    if (group === undefined) {
      throw noSuchObject(id, 'group');
    }
    return group;
  }
  async getAllGroups() {
    return this._groups.get();
  }
  async addUserToGroup(userId, groupId) {
    const [user, group] = await Promise.all([this.getUser(userId), this.getGroup(groupId)]);
    user.groups = addToArraySet(user.groups, groupId);
    group.users = addToArraySet(group.users, userId);
    await Promise.all([this._users.save(user), this._groups.save(group)]);
  }
  async _removeUserFromGroup(userId, group) {
    group.users = removeFromArraySet(group.users, userId);
    return this._groups.save(group);
  }
  async _removeGroupFromUser(groupId, user) {
    user.groups = removeFromArraySet(user.groups, groupId);
    return this._users.save(user);
  }
  async removeUserFromGroup(userId, groupId) {
    const [user, group] = await Promise.all([this.getUser(userId), this.getGroup(groupId)]);
    await Promise.all([this._removeUserFromGroup(userId, group), this._removeGroupFromUser(groupId, user)]);
  }
  async setGroupUsers(groupId, userIds) {
    var _context7;
    const group = await this.getGroup(groupId);
    let newUsersIds = lightSet(userIds);
    const oldUsersIds = [];
    forEach(group.users, id => {
      if (newUsersIds.has(id)) {
        newUsersIds.delete(id);
      } else {
        oldUsersIds.push(id);
      }
    });
    newUsersIds = newUsersIds.toArray();
    const getUser = this.getUser.bind(this);
    const [newUsers, oldUsers] = await Promise.all([Promise.all(newUsersIds.map(getUser)), Promise.all(oldUsersIds.map(getUser))]);
    forEach(newUsers, user => {
      user.groups = addToArraySet(user.groups, groupId);
    });
    forEach(oldUsers, user => {
      user.groups = removeFromArraySet(user.groups, groupId);
    });
    group.users = userIds;
    const saveUser = (_context7 = this._users).save.bind(_context7);
    await Promise.all([Promise.all(newUsers.map(saveUser)), Promise.all(oldUsers.map(saveUser)), this._groups.save(group)]);
  }
}
//# sourceMappingURL=subjects.mjs.map