import assert from 'assert';
import findKey from 'lodash/findKey.js';
import pick from 'lodash/pick.js';
import { asyncEach } from '@vates/async-each';
import { BaseError } from 'make-error';
import { createLogger } from '@xen-orchestra/log';
import { fibonacci } from 'iterable-backoff';
import { networkInterfaces } from 'os';
import { noSuchObject } from 'xo-common/api-errors.js';
import { parseDuration } from '@vates/parse-duration';
import { pDelay, ignoreErrors } from 'promise-toolbox';
import * as XenStore from '../_XenStore.mjs';
import Xapi from '../xapi/index.mjs';
import xapiObjectToXo from '../xapi-object-to-xo.mjs';
import XapiStats from '../xapi-stats.mjs';
import { camelToSnakeCase, forEach, isEmpty, popProperty } from '../utils.mjs';
import { Servers } from '../models/server.mjs';
class PoolAlreadyConnected extends BaseError {
  constructor(poolId, connectedServerId, connectingServerId) {
    super('this pool is already connected');
    this.poolId = poolId;
    this.connectedServerId = connectedServerId;
    this.connectingServerId = connectingServerId;
  }
}
const log = createLogger('xo:xo-mixins:xen-servers');
export default class {
  constructor(app, {
    safeMode
  }) {
    this._objectConflicts = {
      __proto__: null
    };
    const serversDb = this._servers = new Servers({
      connection: app._redis,
      namespace: 'server',
      indexes: ['host']
    });
    this._serverIdsByPool = {
      __proto__: null
    };
    this._stats = new XapiStats();
    this._xapis = {
      __proto__: null
    };
    this._app = app;
    app.config.watchDuration('xapiMarkDisconnectedDelay', xapiMarkDisconnectedDelay => {
      this._xapiMarkDisconnectedDelay = xapiMarkDisconnectedDelay;
    });
    app.hooks.on('clean', () => serversDb.rebuildIndexes());
    app.hooks.on('start', async () => {
      const connectServers = async () => {
        for (const server of await serversDb.get()) {
          if (server.enabled) {
            this.connectXenServer(server.id).catch(error => {
              log.warn('failed to connect to XenServer', {
                host: server.host,
                error
              });
            });
          }
        }
      };
      app.addConfigManager('xenServers', () => serversDb.get(), servers => serversDb.update(servers).then(connectServers));
      if (!(await serversDb.exists())) {
        const key = 'vm-data/xen-servers';
        const xenStoreServers = await XenStore.read(key).then(JSON.parse).catch(() => []);
        for (const server of xenStoreServers) {
          await this.registerXenServer(server);
        }
        ignoreErrors.call(XenStore.rm(key));
      }
      if (!safeMode) {
        await connectServers();
      }
    });
    this._applianceUuid = Math.random().toString(36).slice(2);
    app.hooks.on('start', async () => {
      const value = (await XenStore.read('vm')).trim();
      assert(value.startsWith('/vm/'));
      this._applianceUuid = value.slice(4);
    });
  }
  async registerXenServer({
    allowUnauthorized = false,
    host,
    label,
    password,
    readOnly = false,
    username,
    httpProxy
  }) {
    const server = await this._servers.create({
      allowUnauthorized,
      enabled: true,
      host,
      httpProxy,
      label,
      password,
      readOnly,
      username
    });
    return server;
  }
  async unregisterXenServer(id) {
    var _context;
    (_context = this.disconnectXenServer(id), ignoreErrors).call(_context);
    if (!(await this._servers.remove(id))) {
      throw noSuchObject(id, 'xenServer');
    }
  }
  async updateXenServer(id, {
    allowUnauthorized,
    enabled,
    error,
    host,
    label,
    password,
    readOnly,
    username,
    httpProxy
  }) {
    const server = await this._getXenServer(id);
    const xapi = this._xapis[id];
    const requireDisconnected = allowUnauthorized !== undefined || host !== undefined || password !== undefined || username !== undefined || httpProxy !== undefined;
    if (requireDisconnected && xapi !== undefined && xapi.status !== 'disconnected') {
      throw new Error('this entry require disconnecting the server to update it');
    }
    if (label !== undefined) server.label = label || undefined;
    if (host) server.host = host;
    if (username) server.username = username;
    if (password) server.password = password;
    if (error !== undefined) {
      server.error = error;
    }
    if (enabled !== undefined) {
      server.enabled = enabled;
    }
    if (readOnly !== undefined) {
      server.readOnly = readOnly;
      if (xapi !== undefined) {
        xapi.readOnly = readOnly;
      }
    }
    if (allowUnauthorized !== undefined) {
      server.allowUnauthorized = allowUnauthorized;
    }
    if (httpProxy !== undefined) {
      server.httpProxy = httpProxy === null ? undefined : httpProxy;
    }
    await this._servers.update(server);
  }
  async getXenServer(id) {
    return await this._getXenServer(id);
  }
  async _getXenServer(id) {
    const server = await this._servers.first(id);
    if (server === undefined) {
      throw noSuchObject(id, 'xenServer');
    }
    return server;
  }
  getXenServerIdByObject(object, type) {
    if (typeof object === 'string') {
      object = this._app.getObject(object, type);
    }
    const {
      $pool: poolId
    } = object;
    if (!poolId) {
      throw new Error(`object ${object.id} does not belong to a pool`);
    }
    const serverId = this._serverIdsByPool[poolId];
    if (serverId === undefined) {
      throw new Error(`no connection found for object ${object.id}`);
    }
    return serverId;
  }
  _onXenAdd(newXapiObjects, xapiIdsToXo, toRetry, conId, dependents, xapiObjects) {
    const conflicts = this._objectConflicts;
    const objects = this._app._objects;
    const serverIdsByPool = this._serverIdsByPool;
    forEach(newXapiObjects, function handleObject(xapiObject, xapiId) {
      if (xapiObject.$type === 'pool' && serverIdsByPool[xapiObject.$id] === undefined) {
        const obsoletePoolId = findKey(serverIdsByPool, serverId => serverId === conId);
        delete serverIdsByPool[obsoletePoolId];
        serverIdsByPool[xapiObject.$id] = conId;
      }
      const {
        $ref
      } = xapiObject;
      const dependent = dependents[$ref];
      if (dependent !== undefined) {
        delete dependents[$ref];
        return handleObject(xapiObjects[dependent], dependent);
      }
      try {
        const xoObject = xapiObjectToXo(xapiObject, dependents);
        if (!xoObject) {
          return;
        }
        const xoId = xoObject.id;
        xapiIdsToXo[xapiId] = xoId;
        const previous = objects.get(xoId, undefined);
        if (previous && previous._xapiRef !== $ref) {
          const conflicts_ = conflicts[xoId] || (conflicts[xoId] = {
            __proto__: null
          });
          conflicts_[conId] = xoObject;
        } else {
          objects.set(xoId, xoObject);
        }
      } catch (error) {
        if (xapiId in toRetry) {
          const {
            $pool,
            $ref,
            $type,
            uuid
          } = xapiObject;
          log.error('xapiObjectToXo', {
            error,
            xapiObject: {
              $pool: pick($pool, ['$ref', 'uuid']),
              $ref,
              $type,
              uuid
            }
          });
        }
        toRetry[xapiId] = xapiObject;
      }
    });
  }
  _onXenRemove(xapiObjects, xapiIdsToXo, toRetry, conId) {
    const conflicts = this._objectConflicts;
    const objects = this._app._objects;
    forEach(xapiObjects, (_, xapiId) => {
      toRetry && delete toRetry[xapiId];
      const xoId = xapiIdsToXo[xapiId];
      if (!xoId) {
        return;
      }
      delete xapiIdsToXo[xapiId];
      const objConflicts = conflicts[xoId];
      if (objConflicts) {
        if (objConflicts[conId]) {
          delete objConflicts[conId];
        } else {
          objects.set(xoId, popProperty(objConflicts));
        }
        if (isEmpty(objConflicts)) {
          delete conflicts[xoId];
        }
      } else {
        objects.unset(xoId);
      }
    });
  }
  async connectXenServer(id) {
    const server = await this.getXenServer(id);
    if (this._getXenServerStatus(id) !== 'disconnected') {
      throw new Error('the server is already connected');
    }
    const {
      config
    } = this._app;
    let {
      poolMarkingInterval,
      poolMarkingMaxAge,
      poolMarkingPrefix,
      ...xapiOptions
    } = config.get('xapiOptions');
    poolMarkingInterval = parseDuration(poolMarkingInterval);
    poolMarkingMaxAge = parseDuration(poolMarkingMaxAge);
    const xapi = this._xapis[server.id] = new Xapi({
      allowUnauthorized: server.allowUnauthorized,
      readOnly: server.readOnly,
      ...xapiOptions,
      httpProxy: server.httpProxy,
      guessVhdSizeOnImport: config.get('guessVhdSizeOnImport'),
      auth: {
        user: server.username,
        password: server.password
      },
      url: server.host,
      watchEvents: false
    });
    try {
      var _context3;
      await xapi.connect();
      if (this._xapis[server.id] === undefined) {
        var _context2;
        (_context2 = xapi.disconnect(), ignoreErrors).call(_context2);
        return;
      }
      const serverIdsByPool = this._serverIdsByPool;
      const poolId = xapi.pool.$id;
      if (serverIdsByPool[poolId] !== undefined) {
        throw new PoolAlreadyConnected(poolId, serverIdsByPool[poolId], server.id);
      }
      serverIdsByPool[poolId] = server.id;
      xapi.xo = (() => {
        const conId = server.id;
        const xapiIdsToXo = {
          __proto__: null
        };
        let toRetry;
        let toRetryNext = {
          __proto__: null
        };
        const dependents = {
          __proto__: null
        };
        const onAddOrUpdate = objects => {
          this._onXenAdd(objects, xapiIdsToXo, toRetryNext, conId, dependents, xapi.objects.all);
        };
        const onRemove = objects => {
          this._onXenRemove(objects, xapiIdsToXo, toRetry, conId, dependents);
        };
        const onFinish = () => {
          if (!isEmpty(toRetry)) {
            onAddOrUpdate(toRetry);
            toRetry = null;
          }
          if (!isEmpty(toRetryNext)) {
            toRetry = toRetryNext;
            toRetryNext = {
              __proto__: null
            };
          }
        };
        const {
          objects
        } = xapi;
        const addObject = object => {
          onAddOrUpdate({
            [object.$id]: object
          });
          return xapiObjectToXo(object, dependents);
        };
        const markPool = async () => {
          const now = Date.now();
          const {
            pool
          } = xapi;
          try {
            await asyncEach(Object.entries(pool.other_config), ([key, value]) => {
              if (key.startsWith(poolMarkingPrefix)) {
                const {
                  lastConnected
                } = JSON.parse(value);
                if (now - lastConnected > poolMarkingMaxAge) {
                  return pool.update_other_config(key, null);
                }
              }
            });
            const publicUrl = config.getOptional('http.publicUrl');
            const info = {
              lastConnected: now,
              publicUrl
            };
            if (publicUrl === undefined) {
              info.networkInterfaces = networkInterfaces();
              delete info.networkInterfaces.lo;
            }
            await pool.update_other_config(poolMarkingPrefix + this._applianceUuid, JSON.stringify(info));
          } catch (error) {
            log.warn('markPool', {
              error
            });
          }
        };
        let markPoolHandle;
        return {
          httpRequest: this._app.httpRequest.bind(this),
          install() {
            objects.on('add', onAddOrUpdate);
            objects.on('update', onAddOrUpdate);
            objects.on('remove', onRemove);
            objects.on('finish', onFinish);
            onAddOrUpdate(objects.all);
            markPool();
            markPoolHandle = setInterval(markPool, poolMarkingInterval);
          },
          uninstall() {
            clearInterval(markPoolHandle);
            objects.removeListener('add', onAddOrUpdate);
            objects.removeListener('update', onAddOrUpdate);
            objects.removeListener('remove', onRemove);
            objects.removeListener('finish', onFinish);
            onRemove(objects.all);
          },
          addObject,
          getData: (id, key) => {
            const value = (typeof id === 'string' ? xapi.getObject(id) : id).other_config[`xo:${camelToSnakeCase(key)}`];
            return value && JSON.parse(value);
          },
          setData: async (id, key, value) => {
            key = `xo:${camelToSnakeCase(key)}`;
            value = value !== null ? JSON.stringify(value) : value;
            const object = await xapi.getObject(id);
            if (object.other_config[key] === (value === null ? undefined : value)) {
              return;
            }
            await object.update_other_config(key, value);
            xapi.waitObject(id, addObject);
          }
        };
      })();
      xapi.xo.install();
      xapi.watchEvents();
      (_context3 = this.updateXenServer(id, {
        error: null
      }), ignoreErrors).call(_context3);
      xapi.once('eventFetchingError', function eventFetchingErrorListener() {
        const timeout = setTimeout(() => {
          xapi.xo.uninstall();
          delete serverIdsByPool[poolId];
        }, this._xapiMarkDisconnectedDelay);
        xapi.once('eventFetchingSuccess', () => {
          xapi.once('eventFetchingError', eventFetchingErrorListener);
          if (serverIdsByPool[poolId] === undefined) {
            serverIdsByPool[poolId] = server.id;
            xapi.xo.install();
          } else {
            clearTimeout(timeout);
          }
        });
      });
      xapi.once('disconnected', () => {
        xapi.xo.uninstall();
        delete this._xapis[server.id];
        delete this._serverIdsByPool[poolId];
        this._app.emit('server:disconnected', {
          server,
          xapi
        });
      });
      this._app.emit('server:connected', {
        server,
        xapi
      });
    } catch (error) {
      var _context4;
      delete this._xapis[server.id];
      (_context4 = xapi.disconnect(), ignoreErrors).call(_context4);
      (_context4 = this.updateXenServer(id, {
        error
      }), ignoreErrors).call(_context4);
      throw error;
    }
  }
  async disconnectXenServer(id) {
    const status = this._getXenServerStatus(id);
    if (status === 'disconnected') {
      return;
    }
    const xapi = this._xapis[id];
    delete this._xapis[id];
    const serverIdsByPool = this._serverIdsByPool;
    const poolId = findKey(serverIdsByPool, _ => _ === xapi);
    if (poolId !== undefined) {
      delete serverIdsByPool[id];
    }
    return xapi.disconnect();
  }
  getAllXapis() {
    return this._xapis;
  }
  getXapi(object, type) {
    return this._xapis[this.getXenServerIdByObject(object, type)];
  }
  getXapiObject(xoObjectOrId, xoType) {
    const xoObject = typeof xoObjectOrId === 'string' ? this._app.getObject(xoObjectOrId, xoType) : xoObjectOrId;
    return this.getXapi(xoObject).getObjectByRef(xoObject._xapiRef);
  }
  _getXenServerStatus(id) {
    var _xapi$pool;
    const xapi = this._xapis[id];
    return xapi === undefined ? 'disconnected' : this._serverIdsByPool[(_xapi$pool = xapi.pool) === null || _xapi$pool === void 0 ? void 0 : _xapi$pool.$id] === id ? 'connected' : 'connecting';
  }
  async getAllXenServers() {
    const servers = await this._servers.get();
    const xapis = this._xapis;
    forEach(servers, server => {
      var _xapis$server$id;
      const lastEventFetchedTimestamp = (_xapis$server$id = xapis[server.id]) === null || _xapis$server$id === void 0 ? void 0 : _xapis$server$id.lastEventFetchedTimestamp;
      if (lastEventFetchedTimestamp !== undefined && Date.now() > lastEventFetchedTimestamp + this._xapiMarkDisconnectedDelay) {
        server.error = xapis[server.id].watchEventsError;
      }
      server.status = this._getXenServerStatus(server.id);
      if (server.status === 'connected') {
        server.poolId = xapis[server.id].pool.uuid;
      }
      delete server.password;
    });
    return servers;
  }
  getXapiVmStats(vmId, granularity) {
    return this._stats.getVmStats(this.getXapi(vmId), vmId, granularity);
  }
  getXapiHostStats(hostId, granularity) {
    return this._stats.getHostStats(this.getXapi(hostId), hostId, granularity);
  }
  getXapiSrStats(srId, granularity) {
    return this._stats.getSrStats(this.getXapi(srId), srId, granularity);
  }
  async mergeXenPools(sourcePoolId, targetPoolId, force = false) {
    var _context5;
    const sourceXapi = this.getXapi(sourcePoolId);
    const {
      _auth: {
        user,
        password
      },
      _url: {
        hostname
      }
    } = this.getXapi(targetPoolId);
    sourceXapi.xo.uninstall();
    try {
      await sourceXapi.joinPool(hostname, user, password, force);
    } catch (e) {
      sourceXapi.xo.install();
      throw e;
    }
    (_context5 = this.unregisterXenServer(this._serverIdsByPool[sourcePoolId]), ignoreErrors).call(_context5);
  }
  async detachHostFromPool(hostId) {
    var _context6;
    const xapi = this.getXapi(hostId);
    const poolId = xapi.pool.$id;
    const {
      address
    } = xapi.getObject(hostId);
    await xapi.ejectHostFromPool(hostId);
    (_context6 = this.getXenServer(this._serverIdsByPool[poolId]).then(async properties => {
      const {
        id
      } = await this.registerXenServer({
        ...properties,
        host: address
      });
      for (const delay of fibonacci().take(5).toMs().map(d => d * 60)) {
        await pDelay(delay);
        try {
          await this.connectXenServer(id);
          break;
        } catch (error) {
          if (!(error instanceof PoolAlreadyConnected) && error.code !== 'EHOSTUNREACH') {
            throw error;
          }
        }
      }
    }), ignoreErrors).call(_context6);
  }
}
//# sourceMappingURL=xen-servers.mjs.map