var _dec, _class;
import Config from '@xen-orchestra/mixins/Config.mjs';
import forEach from 'lodash/forEach.js';
import Hooks from '@xen-orchestra/mixins/Hooks.mjs';
import HttpProxy from '@xen-orchestra/mixins/HttpProxy.mjs';
import includes from 'lodash/includes.js';
import isEmpty from 'lodash/isEmpty.js';
import iteratee from 'lodash/iteratee.js';
import mixin from '@xen-orchestra/mixin';
import mixinLegacy from '@xen-orchestra/mixin/legacy.js';
import stubTrue from 'lodash/stubTrue.js';
import SslCertificate from '@xen-orchestra/mixins/SslCertificate.mjs';
import Tasks from '@xen-orchestra/mixins/Tasks.mjs';
import { Collection as XoCollection } from 'xo-collection';
import { createClient as createRedisClient } from 'redis';
import { createDebounceResource } from '@vates/disposable/debounceResource.js';
import { createLogger } from '@xen-orchestra/log';
import { EventEmitter } from 'events';
import { noSuchObject } from 'xo-common/api-errors.js';
import { parseDuration } from '@vates/parse-duration';
import { UniqueIndex as XoUniqueIndex } from 'xo-collection/unique-index.js';
import mixins from './xo-mixins/index.mjs';
import { generateToken, noop } from './utils.mjs';
const log = createLogger('xo:xo');
let Xo = (_dec = mixinLegacy(Object.values(mixins)), _dec(_class = class Xo extends EventEmitter {
  constructor(opts) {
    super();
    mixin(this, {
      Config,
      Hooks,
      HttpProxy,
      SslCertificate,
      Tasks
    }, [opts]);
    this.hooks.setMaxListeners(0);
    const {
      config
    } = opts;
    this._objects = new XoCollection();
    this._objects.createIndex('byRef', new XoUniqueIndex('_xapiRef'));
    this._httpRequestWatchers = {
      __proto__: null
    };
    {
      const {
        socket: path,
        uri: url
      } = config.redis || {};
      const redis = createRedisClient({
        socket: {
          path
        },
        url
      });
      this._redis = redis;
      this.hooks.on('start core', () => redis.connect());
      this.hooks.on('stop core', () => redis.quit());
    }
    this.hooks.on('start', () => this._watchObjects());
    const debounceResource = createDebounceResource();
    debounceResource.defaultDelay = parseDuration(config.resourceCacheDelay);
    this.hooks.on('stop', debounceResource.flushAll);
    this.debounceResource = debounceResource;
    this.version = opts.appVersion;
  }
  get objects() {
    return this._objects;
  }
  getObject(key, type) {
    const {
      all,
      indexes: {
        byRef
      }
    } = this._objects;
    const obj = all[key] || byRef[key];
    if (!obj) {
      throw noSuchObject(key, type);
    }
    if (type != null && (typeof type === 'string' && type !== obj.type || !includes(type, obj.type))) {
      throw noSuchObject(key, type);
    }
    return obj;
  }
  getObjects({
    filter,
    limit
  } = {}) {
    const {
      all
    } = this._objects;
    if (filter === undefined) {
      if (limit === undefined || limit === Infinity) {
        return all;
      }
      filter = stubTrue;
    } else {
      filter = iteratee(filter);
      if (limit === undefined) {
        limit = Infinity;
      }
    }
    const results = {
      __proto__: null
    };
    for (const id in all) {
      const object = all[id];
      if (filter(object, id, all)) {
        if (limit-- <= 0) {
          break;
        }
        results[id] = object;
      }
    }
    return results;
  }
  _handleHttpRequest(req, res, next) {
    const {
      url
    } = req;
    const {
      _httpRequestWatchers: watchers
    } = this;
    const watcher = watchers[url];
    if (!watcher) {
      next();
      return;
    }
    if (!watcher.persistent) {
      delete watchers[url];
    }
    const {
      fn,
      data
    } = watcher;
    new Promise(resolve => {
      resolve(fn.call(this, req, res, data, next));
    }).then(result => {
      if (result != null) {
        if (typeof result === 'string' || Buffer.isBuffer(result)) {
          res.end(result);
        } else if (typeof result.pipe === 'function') {
          result.pipe(res);
        } else {
          res.end(JSON.stringify(result));
        }
      }
    }, error => {
      log.error('HTTP request error', {
        data,
        error,
        fn: fn.name
      });
      if (!res.headersSent) {
        res.writeHead(500);
        res.write('unknown error');
      }
      res.end();
    });
  }
  async registerHttpRequest(fn, data, {
    suffix = ''
  } = {}) {
    const {
      _httpRequestWatchers: watchers
    } = this;
    let url;
    do {
      url = `/api/${await generateToken()}${suffix}`;
    } while (url in watchers);
    watchers[url] = {
      data,
      fn
    };
    return url;
  }
  async registerHttpRequestHandler(url, fn, {
    data = undefined,
    persistent = true
  } = {}) {
    const {
      _httpRequestWatchers: watchers
    } = this;
    if (url in watchers) {
      throw new Error(`a handler is already registered for ${url}`);
    }
    watchers[url] = {
      data,
      fn,
      persistent
    };
  }
  async unregisterHttpRequestHandler(url) {
    delete this._httpRequestWatchers[url];
  }
  defineProperty(name, value, thisArg = null) {
    if (name in this) {
      throw new Error(`Xo#${name} is already defined`);
    }
    if (typeof value === 'function') {
      value = (value => function () {
        return value.apply(thisArg, arguments);
      })(value);
    }
    Object.defineProperty(this, name, {
      configurable: true,
      value
    });
    let unset = () => {
      delete this[name];
      unset = noop;
    };
    return () => unset();
  }
  defineProperties(props, thisArg) {
    const unsets = [];
    const unset = () => forEach(unsets, unset => unset());
    try {
      forEach(props, (value, name) => {
        unsets.push(this.defineProperty(name, value, thisArg));
      });
    } catch (error) {
      unset();
      throw error;
    }
    return unset;
  }
  _watchObjects() {
    const {
      _objects: objects
    } = this;
    let entered, exited;
    function reset() {
      entered = {
        __proto__: null
      };
      exited = {
        __proto__: null
      };
    }
    reset();
    function onAdd(items) {
      forEach(items, (item, id) => {
        entered[id] = item;
      });
    }
    objects.on('add', onAdd);
    objects.on('update', onAdd);
    objects.on('remove', items => {
      forEach(items, (_, id) => {
        exited[id] = 0;
      });
    });
    objects.on('finish', () => {
      const enteredMessage = !isEmpty(entered) && {
        type: 'enter',
        items: entered
      };
      const exitedMessage = !isEmpty(exited) && {
        type: 'exit',
        items: exited
      };
      if (!enteredMessage && !exitedMessage) {
        return;
      }
      for (const connection of this.apiConnections) {
        if (connection.has('user_id') && connection.notify) {
          if (enteredMessage) {
            connection.notify('all', enteredMessage);
          }
          if (exitedMessage) {
            connection.notify('all', exitedMessage);
          }
        }
      }
      reset();
    });
  }
}) || _class);
export { Xo as default };
//# sourceMappingURL=xo.mjs.map